"""Python Module for building solid regular geometric objects.

Return value are list of vectors, 3 vectors define a facet.

Copyright (c) Berthold Grupp          2005
License: LGPL

Sample code for creating a mesh:
	facets = Cube(3.0, 4.0, 5.0)
	m = Mesh.newMesh()
	m.addFacets(facets)
"""


import math

def Sphere (radius, count):
    """Creates a sphere with a given radius.
    
    bla bla bla
    
    """
    return Ellipsoid(radius, radius, count)

def Ellipsoid (lenX, lenY, count):
    polyline = []
    step = math.pi / count
    i = 0.0
    while (i < math.pi + step / 10.0):
        x = math.cos(i) * lenX
        y = math.sin(i) * lenY 
        polyline.append([x, y])
        i = i + step

    return RotationBody(polyline, count)    

def Cylinder (radius, len, closed, edgelen, count):
    return Cone(radius, radius, len, closed, edgelen, count)

def Cone (radius1, radius2, len, closed, edgelen, count):
    polyline = []
    if (closed):
        try:
            step = radius2 / math.ceil(radius2 / edgelen)
        except ZeroDivisionError:
            pass
        else:
            i = 0.0
            while (i < radius2 - step / 2.0):
                polyline.append([len, i])
                i = i + step

    ct = math.ceil(len / edgelen)        
    step = len / ct
    rstep = (radius1 - radius2) / ct
    i = len;
    r = radius2
    while (i > 0.0 + step / 2.0):
        polyline.append([i, r])
        i = i - step
        r = r + rstep
    polyline.append([0.0, radius1])

    if (closed):
        try:
            step = radius1 / math.ceil(radius1 / edgelen)
        except ZeroDivisionError:
            pass
        else:
            i = radius1 - step
            while (i > 0.0 + step / 2.0):
                polyline.append([0.0, i])
                i = i - step
            polyline.append([0.0, 0.0])
        
    return RotationBody(polyline, count)

def Toroid (radius1, radius2, count):
    polyline = []
    
    step = math.pi * 2.0 / count
    i = -math.pi
    while (i < math.pi +  step / 10.0):
        x = radius1 + math.cos(i) * radius2
        y = radius1 + math.sin(i) * radius2
        polyline.append([x, y])
        i = i + step

    return RotationBody(polyline, count)
         

def RotationBody (polyline, count):
    """Build a rotation body from a given (closed) polyline, rotation axis is the X-Axis.
    
    Parameter: polyline: list of tuple of 2 floats (2d vector)
    
    """
    facets = []

    step = math.pi * 2.0 / count
    i = -math.pi;
    while (i < math.pi - step / 10.0):
        li = i + step
        for j in range(0, len(polyline) - 1):
            v1 = polyline[j]
            v2 = polyline[j+1]
            
            x1 = v1[0]
            y1 = v1[1] * math.cos(i)
            z1 = v1[1] * math.sin(i)
            x2 = v2[0]
            y2 = v2[1] * math.cos(i)
            z2 = v2[1] * math.sin(i)
            x3 = v1[0]
            y3 = v1[1] * math.cos(li)
            z3 = v1[1] * math.sin(li)
            x4 = v2[0]
            y4 = v2[1] * math.cos(li)
            z4 = v2[1] * math.sin(li)

            if (v1[1] != 0.0):            
                facets.append([x1, y1, z1])
                facets.append([x2, y2, z2])
                facets.append([x3, y3, z3])

            if (v2[1] != 0.0):
                facets.append([x2, y2, z2])
                facets.append([x4, y4, z4])
                facets.append([x3, y3, z3])
            
        i = i + step
        
    return facets;

def Cube (lenX, lenY, lenZ):
    hx = lenX / 2.0
    hy = lenY / 2.0
    hz = lenZ / 2.0
    
    facets = []
    
    facets.append([-hx, -hy, -hz])
    facets.append([hx, -hy, -hz])
    facets.append([hx, -hy, hz])

    facets.append([-hx, -hy, -hz])
    facets.append([hx, -hy, hz])
    facets.append([-hx, -hy, hz])

    facets.append([-hx, hy, -hz])
    facets.append([hx, hy, hz])
    facets.append([hx, hy, -hz])

    facets.append([-hx, hy, -hz])
    facets.append([-hx, hy, hz])
    facets.append([hx, hy, hz])

    facets.append([-hx, -hy, -hz])
    facets.append([-hx, hy, hz])
    facets.append([-hx, hy, -hz])

    facets.append([-hx, -hy, -hz])
    facets.append([-hx, -hy, hz])
    facets.append([-hx, hy, hz])

    facets.append([hx, -hy, -hz])
    facets.append([hx, hy, -hz])
    facets.append([hx, hy, hz])

    facets.append([hx, -hy, -hz])
    facets.append([hx, hy, hz])
    facets.append([hx, -hy, hz])

    facets.append([-hx, -hy, -hz])
    facets.append([-hx, hy, -hz])
    facets.append([hx, hy, -hz])

    facets.append([-hx, -hy, -hz])
    facets.append([hx, hy, -hz])
    facets.append([hx, -hy, -hz])

    facets.append([-hx, -hy, hz])
    facets.append([hx, hy, hz])
    facets.append([-hx, hy, hz])

    facets.append([-hx, -hy, hz])
    facets.append([hx, -hy, hz])
    facets.append([hx, hy, hz])
    
    return facets

def FineCube (lenX, lenY, lenZ, edgelen):
    hx = lenX / 2.0
    hy = lenY / 2.0
    hz = lenZ / 2.0
    cx = int(max(lenX / edgelen,1))
    dx = lenX / cx
    cy = int(max(lenY / edgelen,1))
    dy = lenY / cy
    cz = int(max(lenZ / edgelen,1))
    dz = lenZ / cz
    
    facets = []

    # z
    for i in range(0,cx):
        for j in range(0,cy):
            facets.append([-hx+(i+0)*dx, -hy+(j+0)*dy, -hz])
            facets.append([-hx+(i+0)*dx, -hy+(j+1)*dy, -hz])
            facets.append([-hx+(i+1)*dx, -hy+(j+1)*dy, -hz])

            facets.append([-hx+(i+0)*dx, -hy+(j+0)*dy, -hz])
            facets.append([-hx+(i+1)*dx, -hy+(j+1)*dy, -hz])
            facets.append([-hx+(i+1)*dx, -hy+(j+0)*dy, -hz])

            facets.append([-hx+(i+0)*dx, -hy+(j+0)*dy, hz])
            facets.append([-hx+(i+1)*dx, -hy+(j+1)*dy, hz])
            facets.append([-hx+(i+0)*dx, -hy+(j+1)*dy, hz])

            facets.append([-hx+(i+0)*dx, -hy+(j+0)*dy, hz])
            facets.append([-hx+(i+1)*dx, -hy+(j+0)*dy, hz])
            facets.append([-hx+(i+1)*dx, -hy+(j+1)*dy, hz])

    # y
    for i in range(0,cx):
        for j in range(0,cz):
            facets.append([-hx+(i+0)*dx, -hy, -hz+(j+0)*dz])
            facets.append([-hx+(i+1)*dx, -hy, -hz+(j+1)*dz])
            facets.append([-hx+(i+0)*dx, -hy, -hz+(j+1)*dz])

            facets.append([-hx+(i+0)*dx, -hy, -hz+(j+0)*dz])
            facets.append([-hx+(i+1)*dx, -hy, -hz+(j+0)*dz])
            facets.append([-hx+(i+1)*dx, -hy, -hz+(j+1)*dz])

            facets.append([-hx+(i+0)*dx, hy, -hz+(j+0)*dz])
            facets.append([-hx+(i+0)*dx, hy, -hz+(j+1)*dz])
            facets.append([-hx+(i+1)*dx, hy, -hz+(j+1)*dz])

            facets.append([-hx+(i+0)*dx, hy, -hz+(j+0)*dz])
            facets.append([-hx+(i+1)*dx, hy, -hz+(j+1)*dz])
            facets.append([-hx+(i+1)*dx, hy, -hz+(j+0)*dz])

    # x
    for i in range(0,cy):
        for j in range(0,cz):
            facets.append([-hx, -hy+(i+0)*dy, -hz+(j+0)*dz])
            facets.append([-hx, -hy+(i+0)*dy, -hz+(j+1)*dz])
            facets.append([-hx, -hy+(i+1)*dy, -hz+(j+1)*dz])

            facets.append([-hx, -hy+(i+0)*dy, -hz+(j+0)*dz])
            facets.append([-hx, -hy+(i+1)*dy, -hz+(j+1)*dz])
            facets.append([-hx, -hy+(i+1)*dy, -hz+(j+0)*dz])

            facets.append([hx, -hy+(i+0)*dy, -hz+(j+0)*dz])
            facets.append([hx, -hy+(i+1)*dy, -hz+(j+1)*dz])
            facets.append([hx, -hy+(i+0)*dy, -hz+(j+1)*dz])

            facets.append([hx, -hy+(i+0)*dy, -hz+(j+0)*dz])
            facets.append([hx, -hy+(i+1)*dy, -hz+(j+0)*dz])
            facets.append([hx, -hy+(i+1)*dy, -hz+(j+1)*dz])

    return facets

def main ():
    Cylinder (10.0, 20.0, 1, 10, 10)
    
if __name__ == "__main__":
    main()
        


# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 - Bernd Hahnebach <bernd@bimstatik.org>            *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM import tools"
__author__ = "Bernd Hahnebach"
__url__ = "http://www.freecadweb.org"

## @package importToolsFem
#  \ingroup FEM
#  \brief FreeCAD FEM import tools

import FreeCAD
from math import pow, sqrt
import numpy as np


def get_FemMeshObjectMeshGroups(fem_mesh_obj):
    """
        Get mesh groups from mesh. This also throws no exception if there
        is no Groups property at all (e.g. Netgen meshes).
    """
    fem_mesh = fem_mesh_obj.FemMesh
    try:
        gmshgroups = fem_mesh.Groups
    except:
        gmshgroups = ()

    return gmshgroups


def get_FemMeshObjectOrder(fem_mesh_obj):
    """
        Gets element order. Element order counting based on number of nodes on
        edges. Edge with 2 nodes -> linear elements, Edge with 3 nodes ->
        quadratic elements, and so on. No edges in mesh -> not determined.
        (Is this possible? Seems to be a very degenerate case.)
        If there are edges with different number of nodes appearing, return
        list of orders.
    """
    presumable_order = None

    edges = fem_mesh_obj.FemMesh.Edges

    if edges != ():
        edges_length_set = list({len(fem_mesh_obj.FemMesh.getElementNodes(e)) for e in edges})
        # only need set to eliminate double entries

        if len(edges_length_set) == 1:
            presumable_order = edges_length_set[0] - 1
        else:
            presumable_order = [el - 1 for el in edges_length_set]
    else:
        print("Found no edges in mesh: Element order determination does not work without them.")

    return presumable_order


def get_FemMeshObjectDimension(fem_mesh_obj):
    """ Count all entities in an abstract sense, to distinguish which dimension the mesh is
        (i.e. linemesh, facemesh, volumemesh)
    """
    dim = None

    if fem_mesh_obj.FemMesh.Nodes != ():
        dim = 0
    if fem_mesh_obj.FemMesh.Edges != ():
        dim = 1
    if fem_mesh_obj.FemMesh.Faces != ():
        dim = 2
    if fem_mesh_obj.FemMesh.Volumes != ():
        dim = 3

    return dim


def get_FemMeshObjectElementTypes(fem_mesh_obj, remove_zero_element_entries=True):
    """
        Spit out all elements in the mesh with their appropriate dimension.
    """
    FreeCAD_element_names_dims = {
        "Node": 0, "Edge": 1, "Hexa": 3, "Polygon": 2, "Polyhedron": 3,
        "Prism": 3, "Pyramid": 3, "Quadrangle": 2, "Tetra": 3, "Triangle": 2}

    eval_dict = locals()  # to access local variables from eval
    elements_list_with_zero = [(eval("fem_mesh_obj.FemMesh." + s + "Count", eval_dict), s, d) for (s, d) in FreeCAD_element_names_dims.items()]
    # ugly but necessary
    if remove_zero_element_entries:
        elements_list = [(num, s, d) for (num, s, d) in elements_list_with_zero if num > 0]
    else:
        elements_list = elements_list_with_zero

    return elements_list


def get_MaxDimElementFromList(elem_list):
    """
        Gets element with the maximal dimension in the mesh to determine cells.
    """
    elem_list.sort(key=lambda t: t[2])
    return elem_list[-1]


def make_femmesh(mesh_data):
    ''' makes an FreeCAD FEM Mesh object from FEM Mesh data
    '''
    import Fem
    mesh = Fem.FemMesh()
    m = mesh_data
    if ('Nodes' in m) and (len(m['Nodes']) > 0):
        FreeCAD.Console.PrintLog("Found: nodes\n")
        if (
            ('Seg2Elem' in m)
            or ('Seg3Elem' in m)
            or ('Tria3Elem' in m)
            or ('Tria6Elem' in m)
            or ('Quad4Elem' in m)
            or ('Quad8Elem' in m)
            or ('Tetra4Elem' in m)
            or ('Tetra10Elem' in m)
            or ('Penta6Elem' in m)
            or ('Penta15Elem' in m)
            or ('Hexa8Elem' in m)
            or ('Hexa20Elem' in m)
        ):

            nds = m['Nodes']
            FreeCAD.Console.PrintLog("Found: elements\n")
            for i in nds:
                n = nds[i]
                mesh.addNode(n[0], n[1], n[2], i)
            elms_hexa8 = m['Hexa8Elem']
            for i in elms_hexa8:
                e = elms_hexa8[i]
                mesh.addVolume([e[0], e[1], e[2], e[3], e[4], e[5], e[6], e[7]], i)
            elms_penta6 = m['Penta6Elem']
            for i in elms_penta6:
                e = elms_penta6[i]
                mesh.addVolume([e[0], e[1], e[2], e[3], e[4], e[5]], i)
            elms_tetra4 = m['Tetra4Elem']
            for i in elms_tetra4:
                e = elms_tetra4[i]
                mesh.addVolume([e[0], e[1], e[2], e[3]], i)
            elms_tetra10 = m['Tetra10Elem']
            for i in elms_tetra10:
                e = elms_tetra10[i]
                mesh.addVolume([e[0], e[1], e[2], e[3], e[4], e[5], e[6], e[7], e[8], e[9]], i)
            elms_penta15 = m['Penta15Elem']
            for i in elms_penta15:
                e = elms_penta15[i]
                mesh.addVolume([e[0], e[1], e[2], e[3], e[4], e[5], e[6], e[7], e[8], e[9],
                                e[10], e[11], e[12], e[13], e[14]], i)
            elms_hexa20 = m['Hexa20Elem']
            for i in elms_hexa20:
                e = elms_hexa20[i]
                mesh.addVolume([e[0], e[1], e[2], e[3], e[4], e[5], e[6], e[7], e[8], e[9],
                                e[10], e[11], e[12], e[13], e[14], e[15], e[16], e[17], e[18], e[19]], i)
            elms_tria3 = m['Tria3Elem']
            for i in elms_tria3:
                e = elms_tria3[i]
                mesh.addFace([e[0], e[1], e[2]], i)
            elms_tria6 = m['Tria6Elem']
            for i in elms_tria6:
                e = elms_tria6[i]
                mesh.addFace([e[0], e[1], e[2], e[3], e[4], e[5]], i)
            elms_quad4 = m['Quad4Elem']
            for i in elms_quad4:
                e = elms_quad4[i]
                mesh.addFace([e[0], e[1], e[2], e[3]], i)
            elms_quad8 = m['Quad8Elem']
            for i in elms_quad8:
                e = elms_quad8[i]
                mesh.addFace([e[0], e[1], e[2], e[3], e[4], e[5], e[6], e[7]], i)
            elms_seg2 = m['Seg2Elem']
            for i in elms_seg2:
                e = elms_seg2[i]
                mesh.addEdge([e[0], e[1]], i)
            elms_seg3 = m['Seg3Elem']
            for i in elms_seg3:
                e = elms_seg3[i]
                mesh.addEdge([e[0], e[1], e[2]], i)
            FreeCAD.Console.PrintLog("imported mesh: {} nodes, {} HEXA8, {} PENTA6, {} TETRA4, {} TETRA10, {} PENTA15".format(
                len(nds), len(elms_hexa8), len(elms_penta6), len(elms_tetra4), len(elms_tetra10), len(elms_penta15)
            ))
            FreeCAD.Console.PrintLog("imported mesh: {} HEXA20, {} TRIA3, {} TRIA6, {} QUAD4, {} QUAD8, {} SEG2, {} SEG3".format(
                len(elms_hexa20), len(elms_tria3), len(elms_tria6), len(elms_quad4), len(elms_quad8), len(elms_seg2), len(elms_seg3)
            ))
        else:
            FreeCAD.Console.PrintError("No Elements found!\n")
    else:
        FreeCAD.Console.PrintError("No Nodes found!\n")
    return mesh


def fill_femresult_mechanical(res_obj, result_set):
    ''' fills a FreeCAD FEM mechanical result object with result data
    '''
    if 'number' in result_set:
        eigenmode_number = result_set['number']
    else:
        eigenmode_number = 0
    if 'time' in result_set:
        step_time = result_set['time']
        step_time = round(step_time, 2)

    # if disp exists, fill res_obj.NodeNumbers and res_obj.DisplacementVectors as well as stress and strain
    if 'disp' in result_set:
        disp = result_set['disp']
        displacement = []
        for k, v in disp.items():
            displacement.append(v)

        x_max, y_max, z_max = map(max, zip(*displacement))
        if eigenmode_number > 0:
            span = get_span(res_obj.Mesh.FemMesh.Nodes.items())
            max_disp = max(x_max, y_max, z_max)
            # Allow for max displacement to be 0.1% of the span
            # FIXME - add to Preferences
            max_allowed_disp = 0.001 * span
            scale = max_allowed_disp / max_disp
        else:
            scale = 1.0

        res_obj.DisplacementVectors = list(map((lambda x: x * scale), disp.values()))
        res_obj.NodeNumbers = list(disp.keys())

        # fill res_obj.StressVectors if they exist
        if 'stress' in result_set:
            stress = result_set['stress']
            stressv1 = {}
            for i, stval in enumerate(stress.values()):  # i .. stresstuple .. (Sxx, Syy, Szz, Sxy, Syz, Szx)
                stressv1[i] = (FreeCAD.Vector(stval[0], stval[1], stval[2]))  # Sxx, Syy, Szz
            res_obj.StressVectors = list(map((lambda x: x * scale), stressv1.values()))
            stress_keys = list(stress.keys())
            if (res_obj.NodeNumbers != 0 and res_obj.NodeNumbers != stress_keys):
                print("Inconsistent FEM results: element number for Stress doesn't equal element number for Displacement {} != {}"
                      .format(res_obj.NodeNumbers, len(res_obj.StressValues)))

        # fill res_obj.StrainVectors if they exist
        if 'strainv' in result_set:
            strainv = result_set['strainv']
            res_obj.StrainVectors = list(map((lambda x: x * scale), strainv.values()))

        # calculate von Mises, principal and max Shear and fill them in res_obj
        if 'stress' in result_set:
            stress = result_set['stress']
            if len(stress) > 0:
                mstress = []
                prinstress1 = []
                prinstress2 = []
                prinstress3 = []
                shearstress = []
                for i in stress.values():
                    mstress.append(calculate_von_mises(i))
                    prin1, prin2, prin3, shear = calculate_principal_stress(i)
                    prinstress1.append(prin1)
                    prinstress2.append(prin2)
                    prinstress3.append(prin3)
                    shearstress.append(shear)
                if eigenmode_number > 0:
                    res_obj.StressValues = list(map((lambda x: x * scale), mstress))
                    res_obj.PrincipalMax = list(map((lambda x: x * scale), prinstress1))
                    res_obj.PrincipalMed = list(map((lambda x: x * scale), prinstress2))
                    res_obj.PrincipalMin = list(map((lambda x: x * scale), prinstress3))
                    res_obj.MaxShear = list(map((lambda x: x * scale), shearstress))
                    res_obj.Eigenmode = eigenmode_number
                else:
                    res_obj.StressValues = mstress
                    res_obj.PrincipalMax = prinstress1
                    res_obj.PrincipalMed = prinstress2
                    res_obj.PrincipalMin = prinstress3
                    res_obj.MaxShear = shearstress

        # fill Equivalent Plastic strain if they exist
        if 'peeq' in result_set:
            Peeq = result_set['peeq']
            if len(Peeq) > 0:
                if len(Peeq.values()) != len(disp.values()):
                    Pe = []
                    Pe_extra_nodes = list(Peeq.values())
                    nodes = len(disp.values())
                    for i in range(nodes):
                        Pe_value = Pe_extra_nodes[i]
                        Pe.append(Pe_value)
                    res_obj.Peeq = Pe
                else:
                    res_obj.Peeq = list(Peeq.values())

    # fill res_obj.Temperature if they exist
    # TODO, check if it is possible to have Temperature without disp, we would need to set NodeNumbers than
    if 'temp' in result_set:
        Temperature = result_set['temp']
        if len(Temperature) > 0:
            if len(Temperature.values()) != len(disp.values()):
                Temp = []
                Temp_extra_nodes = list(Temperature.values())
                nodes = len(disp.values())
                for i in range(nodes):
                    Temp_value = Temp_extra_nodes[i]
                    Temp.append(Temp_value)
                res_obj.Temperature = list(map((lambda x: x), Temp))
            else:
                res_obj.Temperature = list(map((lambda x: x), Temperature.values()))
            res_obj.Time = step_time

    # fill res_obj.MassFlow
    if 'mflow' in result_set:
        MassFlow = result_set['mflow']
        if len(MassFlow) > 0:
            res_obj.MassFlowRate = list(map((lambda x: x), MassFlow.values()))
            res_obj.Time = step_time
            res_obj.NodeNumbers = list(MassFlow.keys())  # disp does not exist, res_obj.NodeNumbers needs to be set

    # fill res_obj.NetworkPressure, disp does not exist, see MassFlow
    if 'npressure' in result_set:
        NetworkPressure = result_set['npressure']
        if len(NetworkPressure) > 0:
            res_obj.NetworkPressure = list(map((lambda x: x), NetworkPressure.values()))
            res_obj.Time = step_time

    return res_obj


# helper
def calculate_von_mises(i):
    # Von mises stress (http://en.wikipedia.org/wiki/Von_Mises_yield_criterion)
    s11 = i[0]
    s22 = i[1]
    s33 = i[2]
    s12 = i[3]
    s23 = i[4]
    s31 = i[5]
    s11s22 = pow(s11 - s22, 2)
    s22s33 = pow(s22 - s33, 2)
    s33s11 = pow(s33 - s11, 2)
    s12s23s31 = 6 * (pow(s12, 2) + pow(s23, 2) + pow(s31, 2))
    vm_stress = sqrt(0.5 * (s11s22 + s22s33 + s33s11 + s12s23s31))
    return vm_stress


def calculate_principal_stress(i):
    sigma = np.array([[i[0], i[3], i[5]],
                      [i[3], i[1], i[4]],
                      [i[5], i[4], i[2]]])  # https://forum.freecadweb.org/viewtopic.php?f=18&t=24637&start=10#p240408

    try:  # it will fail if NaN is inside the array,
        # compute principal stresses
        eigvals = list(np.linalg.eigvalsh(sigma))
        eigvals.sort()
        eigvals.reverse()
        maxshear = (eigvals[0] - eigvals[2]) / 2.0
        return (eigvals[0], eigvals[1], eigvals[2], maxshear)
    except:
        return (float('NaN'), float('NaN'), float('NaN'), float('NaN'))
    # TODO might be possible without a try except for NaN, https://forum.freecadweb.org/viewtopic.php?f=22&t=33911&start=10#p284229


def get_span(node_items):
    positions = []  # list of node vectors
    for k, v in node_items:
        positions.append(v)
    p_x_max, p_y_max, p_z_max = map(max, zip(*positions))
    p_x_min, p_y_min, p_z_min = map(min, zip(*positions))
    x_span = abs(p_x_max - p_x_min)
    y_span = abs(p_y_max - p_y_min)
    z_span = abs(p_z_max - p_z_min)
    span = max(x_span, y_span, z_span)
    return span



from __future__ import print_function
import datetime
from PathScripts import PostUtils

# ***************************************************************************
# *   (c) sliptonic (shopinthewoods@gmail.com) 2014                         *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/


TOOLTIP='''
This is an postprocessor file for the Path workbench. It will output path data
in a format suitable for OpenSBP controllers like shopbot.  This postprocessor,
once placed in the appropriate PathScripts folder, can be used directly from
inside FreeCAD, via the GUI importer or via python scripts with:

import Path
Path.write(object,"/path/to/file.ncc","post_opensbp")
'''

'''
DONE:
    uses native commands
    handles feed and jog moves
    handles XY, Z, and XYZ feed speeds
    handles arcs
    support for inch output
ToDo
    comments may not format correctly
    drilling.  Haven't looked at it.
    many other things

'''

TOOLTIP_ARGS='''
Arguments for opensbp:
    --comments          ... insert comments - mostly for debugging
    --inches            ... convert output to inches
    --no-header         ... suppress header output
    --no-show-editor    ... don't show editor, just save result
'''

now = datetime.datetime.now()

OUTPUT_COMMENTS = False
OUTPUT_HEADER = True
SHOW_EDITOR = True
COMMAND_SPACE = ","

# Preamble text will appear at the beginning of the GCODE output file.
PREAMBLE = ''''''
# Postamble text will appear following the last operation.
POSTAMBLE = ''''''

# Pre operation text will be inserted before every operation
PRE_OPERATION = ''''''

# Post operation text will be inserted after every operation
POST_OPERATION = ''''''

# Tool Change commands will be inserted before a tool change
TOOL_CHANGE = ''''''

# to distinguish python built-in open function from the one declared below
if open.__module__ in ['__builtin__','io']:
    pythonopen = open

CurrentState = {}

def getMetricValue(val):
    return val
def getImperialValue(val):
    return val / 25.4

GetValue = getMetricValue


def export(objectslist, filename, argstring):
    global OUTPUT_COMMENTS
    global OUTPUT_HEADER
    global SHOW_EDITOR
    global CurrentState
    global GetValue

    for arg in argstring.split():
        if arg == '--comments':
            OUTPUT_COMMENTS = True
        if arg == '--inches':
            GetValue = getImperialValue
        if arg == '--no-header':
            OUTPUT_HEADER = False
        if arg == '--no-show-editor':
            SHOW_EDITOR = False


    for obj in objectslist:
        if not hasattr(obj, "Path"):
            s = "the object " + obj.Name
            s += " is not a path. Please select only path and Compounds."
            print(s)
            return

    CurrentState = {
        'X': 0, 'Y': 0, 'Z': 0, 'F': 0, 'S': 0,
        'JSXY': 0, 'JSZ': 0, 'MSXY': 0, 'MSZ': 0
    }
    print("postprocessing...")
    gcode = ""

    # write header
    if OUTPUT_HEADER:
        gcode += linenumber() + "'Exported by FreeCAD\n"
        gcode += linenumber() + "'Post Processor: " + __name__ + "\n"
        gcode += linenumber() + "'Output Time:" + str(now) + "\n"

    # Write the preamble
    if OUTPUT_COMMENTS:
        gcode += linenumber() + "'(begin preamble)\n"
    for line in PREAMBLE.splitlines(True):
        gcode += linenumber() + line

    for obj in objectslist:

        # do the pre_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "'(begin operation: " + obj.Label + ")\n"
        for line in PRE_OPERATION.splitlines(True):
            gcode += linenumber() + line

        gcode += parse(obj)

        # do the post_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "'(finish operation: " + obj.Label + ")\n"
        for line in POST_OPERATION.splitlines(True):
            gcode += linenumber() + line

    # do the post_amble
    if OUTPUT_COMMENTS:
        gcode += "'(begin postamble)\n"
    for line in POSTAMBLE.splitlines(True):
        gcode += linenumber() + line

    if SHOW_EDITOR:
        dia = PostUtils.GCodeEditorDialog()
        dia.editor.setText(gcode)
        result = dia.exec_()
        if result:
            final = dia.editor.toPlainText()
        else:
            final = gcode
    else:
        final = gcode

    print("done postprocessing.")

    # Write the output
    gfile = pythonopen(filename, "w")
    gfile.write(final)
    gfile.close()


def move(command):
    global CurrentState

    txt = ""

    # if 'F' in command.Parameters:
    #     txt += feedrate(command)

    axis = ""
    for p in ['X', 'Y', 'Z']:
        if p in command.Parameters:
            if command.Parameters[p] != CurrentState[p]:
                axis += p

    if 'F' in command.Parameters:
        speed = command.Parameters['F']
        if command.Name in ['G1', 'G01']:  # move
            movetype = "MS"
        else:  # jog
            movetype = "JS"
        zspeed = ""
        xyspeed = ""
        if 'Z' in axis:
            speedKey = "{}Z".format(movetype)
            speedVal = GetValue(speed)
            if CurrentState[speedKey] != speedVal:
                CurrentState[speedKey] = speedVal
                zspeed = "{:f}".format(speedVal)
        if ('X' in axis) or ('Y' in axis):
            speedKey = "{}XY".format(movetype)
            speedVal = GetValue(speed)
            if CurrentState[speedKey] != speedVal:
                CurrentState[speedKey] = speedVal
                xyspeed = "{:f}".format(speedVal)
        if zspeed or xyspeed:
            txt += "{},{},{}\n".format(movetype, xyspeed, zspeed)

    if command.Name in ['G0', 'G00']:
        pref = "J"
    else:
        pref = "M"

    if axis == "X":
        txt += pref + "X"
        txt += "," + format(GetValue(command.Parameters["X"]), '.4f')
        txt += "\n"
    elif axis == "Y":
        txt += pref + "Y"
        txt += "," + format(GetValue(command.Parameters["Y"]), '.4f')
        txt += "\n"
    elif axis == "Z":
        txt += pref + "Z"
        txt += "," + format(GetValue(command.Parameters["Z"]), '.4f')
        txt += "\n"
    elif axis == "XY":
        txt += pref + "2"
        txt += "," + format(GetValue(command.Parameters["X"]), '.4f')
        txt += "," + format(GetValue(command.Parameters["Y"]), '.4f')
        txt += "\n"
    elif axis == "XZ":
        txt += pref + "3"
        txt += "," + format(GetValue(command.Parameters["X"]), '.4f')
        txt += ","
        txt += "," + format(GetValue(command.Parameters["Z"]), '.4f')
        txt += "\n"
    elif axis == "XYZ":
        txt += pref + "3"
        txt += "," + format(GetValue(command.Parameters["X"]), '.4f')
        txt += "," + format(GetValue(command.Parameters["Y"]), '.4f')
        txt += "," + format(GetValue(command.Parameters["Z"]), '.4f')
        txt += "\n"
    elif axis == "YZ":
        txt += pref + "3"
        txt += ","
        txt += "," + format(GetValue(command.Parameters["Y"]), '.4f')
        txt += "," + format(GetValue(command.Parameters["Z"]), '.4f')
        txt += "\n"
    elif axis == "":
        print("warning: skipping duplicate move.")
    else:
        print(CurrentState)
        print(command)
        print("I don't know how to handle '{}' for a move.".format(axis))

    return txt


def arc(command):
    if command.Name == 'G2':  # CW
        dirstring = "1"
    else:  # G3 means CCW
        dirstring = "-1"
    txt = "CG,,"
    txt += format(GetValue(command.Parameters['X']), '.4f') + ","
    txt += format(GetValue(command.Parameters['Y']), '.4f') + ","
    txt += format(GetValue(command.Parameters['I']), '.4f') + ","
    txt += format(GetValue(command.Parameters['J']), '.4f') + ","
    txt += "T" + ","
    txt += dirstring
    txt += "\n"
    return txt


def tool_change(command):
    txt = ""
    if OUTPUT_COMMENTS:
        txt += "'a tool change happens now\n"
    for line in TOOL_CHANGE.splitlines(True):
        txt += line
    txt += "&ToolName=" + str(int(command.Parameters['T']))
    txt += "\n"
    txt += "&Tool=" + str(int(command.Parameters['T']))
    txt += "\n"
    return txt


def comment(command):
    print("a comment")
    return


def spindle(command):
    txt = ""
    if command.Name == "M3":  # CW
        pass
    else:
        pass
    txt += "TR," + str(command.Parameters['S']) + "\n"
    txt += "C6\n"
    txt += "PAUSE 2\n"
    return txt


# Supported Commands
scommands = {
    "G0": move,
    "G1": move,
    "G2": arc,
    "G3": arc,
    "M6": tool_change,
    "M3": spindle,
    "G00": move,
    "G01": move,
    "G02": arc,
    "G03": arc,
    "M06": tool_change,
    "M03": spindle,
    "message": comment
}


def parse(pathobj):
    global CurrentState

    output = ""
    params = ['X', 'Y', 'Z', 'A', 'B', 'I', 'J', 'K', 'F', 'S', 'T']
    # Above list controls the order of parameters

    if hasattr(pathobj, "Group"):  # We have a compound or project.
        if OUTPUT_COMMENTS:
            output += linenumber() + "'(compound: " + pathobj.Label + ")\n"
        for p in pathobj.Group:
            output += parse(p)
    else:  # parsing simple path
        # groups might contain non-path things like stock.
        if not hasattr(pathobj, "Path"):
            return output
        if OUTPUT_COMMENTS:
            output += linenumber() + "'(Path: " + pathobj.Label + ")\n"
        for c in pathobj.Path.Commands:
            command = c.Name
            if command in scommands:
                output += scommands[command](c)
                if c.Parameters:
                    CurrentState.update(c.Parameters)
            elif command[0] == '(':
                output += "' " + command + "\n"
            else:
                print("I don't know what the hell the command: ",end='')
                print(command + " means.  Maybe I should support it.")
    return output


def linenumber():
    return ""


print(__name__ + " gcode postprocessor loaded.")



# -*- coding: utf-8 -*-

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2016 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import PathScripts
import math
import unittest

from FreeCAD import Vector
from PathScripts.PathDressupHoldingTags import *
from PathTests.PathTestUtils import PathTestBase

class TestHoldingTags(PathTestBase):
    """Unit tests for the HoldingTags dressup."""

    def test00(self):
        """Check Tag origin."""
        tag = Tag(0, 77, 13, 4, 5, 90, True)
        self.assertCoincide(tag.originAt(3), Vector(77, 13, 3))


    def test01(self):
        """Verify solid for a 90 degree tag is a cylinder."""
        tag = Tag(0, 100, 200, 4, 5, 90, 0, True)
        tag.createSolidsAt(17, 0)

        self.assertIsNotNone(tag.solid)
        self.assertCylinderAt(tag.solid, Vector(100, 200, 17 - 5 * 0.01), 2, 5 * 1.01)

    def test02(self):
        """Verify trapezoidal tag has a cone shape with a lid."""
        tag = Tag(0, 0, 0, 18, 5, 45, 0, True)
        tag.createSolidsAt(0, 0)

        self.assertIsNotNone(tag.solid)
        self.assertConeAt(tag.solid, Vector(0,0,-0.05), 9, 3.95, 5.05)

    def test03(self):
        """Verify pointy cone shape of tag with pointy end if width, angle and height match up."""
        tag = Tag(0, 0, 0, 10, 5, 45, 0, True)
        tag.createSolidsAt(0, 0)
        self.assertIsNotNone(tag.solid)
        h = 5 * 1.01
        self.assertConeAt(tag.solid, Vector(0,0,-h * 0.01), 5, 0, h)

    def test04(self):
        """Verify height adjustment if tag isn't wide eough for angle."""
        tag = Tag(0, 0, 0, 5, 17, 60, 0, True)
        tag.createSolidsAt(0, 0)
        self.assertIsNotNone(tag.solid)
        h = 2.5 * math.tan((60/180.0)*math.pi) * 1.01
        print(h)
        self.assertConeAt(tag.solid, Vector(0,0,-h * 0.01), 2.5, 0, h)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# test script for TechDraw module
# creates a page, 1 view and 1 section view
from __future__ import print_function

import FreeCAD
import Part
import Measure
import TechDraw
import os

def DVSectionTest():
    path = os.path.dirname(os.path.abspath(__file__))
    print ('TDSection path: ' + path)
    templateFileSpec = path + '/TestTemplate.svg'

    FreeCAD.newDocument("TDSection")
    FreeCAD.setActiveDocument("TDSection")
    FreeCAD.ActiveDocument=FreeCAD.getDocument("TDSection")

    box = FreeCAD.ActiveDocument.addObject("Part::Box","Box")

    page = FreeCAD.ActiveDocument.addObject('TechDraw::DrawPage','Page')
    FreeCAD.ActiveDocument.addObject('TechDraw::DrawSVGTemplate','Template')
    FreeCAD.ActiveDocument.Template.Template = templateFileSpec
    FreeCAD.ActiveDocument.Page.Template = FreeCAD.ActiveDocument.Template
    page.Scale = 5.0
#    page.ViewObject.show()    # unit tests run in console mode
    print("page created")

    view = FreeCAD.ActiveDocument.addObject('TechDraw::DrawViewPart','View')
    rc = page.addView(view)
    view.Source = [box]
    view.Direction = (0.0,0.0,1.0)
    view.Rotation = 0.0
    view.X = 30.0
    view.Y = 150.0
    print("view created")

    section = FreeCAD.ActiveDocument.addObject('TechDraw::DrawViewSection','Section')
    rc = page.addView(section)
    section.Source = [box]
    section.BaseView = view
    section.Direction = (0.0,1.0,0.0)
    section.SectionNormal = (0.0,1.0,0.0)
    section.SectionOrigin = (5.0,5.0,5.0)
    view.touch()
    print("section created")

    FreeCAD.ActiveDocument.recompute()
    rc = False
    if ("Up-to-date" in view.State) and ("Up-to-date" in section.State):
        rc = True
    FreeCAD.closeDocument("TDSection")
    return rc

if __name__ == '__main__':
    DVSectionTest()



# -*- coding: utf-8 -*-

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Pekka Roivainen <pekkaroi@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import FreeCAD
import FreeCADGui
import Path
import Part
import PathScripts.PathDressup as PathDressup
import PathScripts.PathGeom as PathGeom
import PathScripts.PathLog as PathLog
import math

from PathScripts import PathUtils
from PySide import QtCore


# Qt tanslation handling
def translate(text, context="Path_DressupRampEntry", disambig=None):
    return QtCore.QCoreApplication.translate(context, text, disambig)


PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())


class ObjectDressup:

    def __init__(self, obj):
        self.obj = obj
        obj.addProperty("App::PropertyLink", "Base", "Path", QtCore.QT_TRANSLATE_NOOP("Path_DressupRampEntry", "The base path to modify"))
        obj.addProperty("App::PropertyAngle", "Angle", "Path", QtCore.QT_TRANSLATE_NOOP("Path_DressupRampEntry", "Angle of ramp."))
        obj.addProperty("App::PropertyEnumeration", "Method", "Path", QtCore.QT_TRANSLATE_NOOP("App::Property", "Ramping Method"))
        obj.addProperty("App::PropertyEnumeration", "RampFeedRate", "FeedRate", QtCore.QT_TRANSLATE_NOOP("App::Property", "Which feed rate to use for ramping"))
        obj.addProperty("App::PropertySpeed", "CustomFeedRate", "FeedRate", QtCore.QT_TRANSLATE_NOOP("App::Property", "Custom feedrate"))
        obj.addProperty("App::PropertyBool", "UseStartDepth", "StartDepth", QtCore.QT_TRANSLATE_NOOP("App::Property", "Should the dressup ignore motion commands above DressupStartDepth"))
        obj.addProperty("App::PropertyDistance", "DressupStartDepth", "StartDepth", QtCore.QT_TRANSLATE_NOOP("App::Property", "The depth where the ramp dressup is enabled. Above this ramps are not generated, but motion commands are passed through as is."))
        obj.Method = ['RampMethod1', 'RampMethod2', 'RampMethod3', 'Helix']
        obj.RampFeedRate = ['Horizontal Feed Rate', 'Vertical Feed Rate', 'Custom']
        obj.Proxy = self
        self.setEditorProperties(obj)

    def __getstate__(self):
        return None

    def __setstate__(self, state):
        return None

    def onChanged(self, obj, prop):
        if prop in ["RampFeedRate", "UseStartDepth"]:
            self.setEditorProperties(obj)

    def setEditorProperties(self, obj):
        if hasattr(obj, 'UseStartDepth'):
            if obj.UseStartDepth:
                obj.setEditorMode('DressupStartDepth', 0)
            else:
                obj.setEditorMode('DressupStartDepth', 2)

        if obj.RampFeedRate == 'Custom':
            obj.setEditorMode('CustomFeedRate', 0)
        else:
            obj.setEditorMode('CustomFeedRate', 2)

    def onDocumentRestored(self, obj):
        self.setEditorProperties(obj)

    def setup(self, obj):
        obj.Angle = 60
        obj.Method = 2
        if PathDressup.baseOp(obj).StartDepth is not None:
            obj.DressupStartDepth = PathDressup.baseOp(obj).StartDepth

    def execute(self, obj):
        if not obj.Base:
            return
        if not obj.Base.isDerivedFrom("Path::Feature"):
            return
        if not obj.Base.Path:
            return
        if obj.Angle >= 90:
            obj.Angle = 89.9
        elif obj.Angle <= 0:
            obj.Angle = 0.1

        if hasattr(obj, 'UseStartDepth'):
            self.ignoreAboveEnabled = obj.UseStartDepth
            self.ignoreAbove = obj.DressupStartDepth
        else:
            self.ignoreAboveEnabled = False
            self.ignoreAbove = 0

        self.angle = obj.Angle
        self.method = obj.Method
        self.wire, self.rapids = PathGeom.wireForPath(obj.Base.Path)
        if self.method in ['RampMethod1', 'RampMethod2', 'RampMethod3']:
            self.outedges = self.generateRamps()
        else:
            self.outedges = self.generateHelix()
        obj.Path = self.createCommands(obj, self.outedges)

    def generateRamps(self, allowBounce=True):
        edges = self.wire.Edges
        outedges = []
        for edge in edges:
            israpid = False
            for redge in self.rapids:
                if PathGeom.edgesMatch(edge, redge):
                    israpid = True
            if not israpid:
                bb = edge.BoundBox
                p0 = edge.Vertexes[0].Point
                p1 = edge.Vertexes[1].Point
                rampangle = self.angle
                if bb.XLength < 1e-6 and bb.YLength < 1e-6 and bb.ZLength > 0 and p0.z > p1.z:

                    # check if above ignoreAbove parameter - do not generate ramp if it is
                    newEdge, cont = self.checkIgnoreAbove(edge)
                    if newEdge is not None:
                        outedges.append(newEdge)
                        p0.z = self.ignoreAbove
                    if cont:
                        continue

                    plungelen = abs(p0.z - p1.z)
                    projectionlen = plungelen * math.tan(math.radians(rampangle))  # length of the forthcoming ramp projected to XY plane
                    PathLog.debug("Found plunge move at X:{} Y:{} From Z:{} to Z{}, length of ramp: {}".format(p0.x, p0.y, p0.z, p1.z, projectionlen))
                    if self.method == 'RampMethod3':
                        projectionlen = projectionlen / 2

                    # next need to determine how many edges in the path after
                    # plunge are needed to cover the length:
                    covered = False
                    coveredlen = 0
                    rampedges = []
                    i = edges.index(edge) + 1
                    while not covered:
                        candidate = edges[i]
                        cp0 = candidate.Vertexes[0].Point
                        cp1 = candidate.Vertexes[1].Point
                        if abs(cp0.z - cp1.z) > 1e-6:
                            # this edge is not parallel to XY plane, not qualified for ramping.
                            break
                        # PathLog.debug("Next edge length {}".format(candidate.Length))
                        rampedges.append(candidate)
                        coveredlen = coveredlen + candidate.Length

                        if coveredlen > projectionlen:
                            covered = True
                        i = i + 1
                        if i >= len(edges):
                            break
                    if len(rampedges) == 0:
                        PathLog.debug("No suitable edges for ramping, plunge will remain as such")
                        outedges.append(edge)
                    else:
                        if not covered:
                            if (not allowBounce) or self.method == 'RampMethod2':
                                l = 0
                                for redge in rampedges:
                                    l = l + redge.Length
                                if self.method == 'RampMethod3':
                                    rampangle = math.degrees(math.atan(l / (plungelen / 2)))
                                else:
                                    rampangle = math.degrees(math.atan(l / plungelen))
                                PathLog.warning("Cannot cover with desired angle, tightening angle to: {}".format(rampangle))

                        # PathLog.debug("Doing ramp to edges: {}".format(rampedges))
                        if self.method == 'RampMethod1':
                            outedges.extend(self.createRampMethod1(rampedges, p0, projectionlen, rampangle))
                        elif self.method == 'RampMethod2':
                            outedges.extend(self.createRampMethod2(rampedges, p0, projectionlen, rampangle))
                        else:
                            # if the ramp cannot be covered with Method3, revert to Method1
                            # because Method1 support going back-and-forth and thus results in same path as Method3 when
                            # length of the ramp is smaller than needed for single ramp.
                            if (not covered) and allowBounce:
                                projectionlen = projectionlen * 2
                                outedges.extend(self.createRampMethod1(rampedges, p0, projectionlen, rampangle))
                            else:
                                outedges.extend(self.createRampMethod3(rampedges, p0, projectionlen, rampangle))
                else:
                    outedges.append(edge)
            else:
                outedges.append(edge)
        return outedges

    def generateHelix(self):
        edges = self.wire.Edges
        minZ = self.findMinZ(edges)
        PathLog.debug("Minimum Z in this path is {}".format(minZ))
        outedges = []
        i = 0
        while i < len(edges):
            edge = edges[i]
            israpid = False
            for redge in self.rapids:
                if PathGeom.edgesMatch(edge, redge):
                    israpid = True
            if not israpid:
                bb = edge.BoundBox
                p0 = edge.Vertexes[0].Point
                p1 = edge.Vertexes[1].Point
                if bb.XLength < 1e-6 and bb.YLength < 1e-6 and bb.ZLength > 0 and p0.z > p1.z:
                    # plungelen = abs(p0.z-p1.z)
                    PathLog.debug("Found plunge move at X:{} Y:{} From Z:{} to Z{}, Searching for closed loop".format(p0.x, p0.y, p0.z, p1.z))
                    # check if above ignoreAbove parameter - do not generate helix if it is
                    newEdge, cont = self.checkIgnoreAbove(edge)
                    if newEdge is not None:
                        outedges.append(newEdge)
                        p0.z = self.ignoreAbove
                    if cont:
                        i = i + 1
                        continue
                    # next need to determine how many edges in the path after plunge are needed to cover the length:
                    loopFound = False
                    rampedges = []
                    j = i + 1
                    while not loopFound:
                        candidate = edges[j]
                        cp0 = candidate.Vertexes[0].Point
                        cp1 = candidate.Vertexes[1].Point
                        if PathGeom.pointsCoincide(p1, cp1):
                            # found closed loop
                            loopFound = True
                            rampedges.append(candidate)
                            break
                        if abs(cp0.z - cp1.z) > 1e-6:
                            # this edge is not parallel to XY plane, not qualified for ramping.
                            break
                        # PathLog.debug("Next edge length {}".format(candidate.Length))
                        rampedges.append(candidate)
                        j = j + 1
                        if j >= len(edges):
                            break
                    if len(rampedges) == 0 or not loopFound:
                        PathLog.debug("No suitable helix found")
                        outedges.append(edge)
                    else:
                        outedges.extend(self.createHelix(rampedges, p0, p1))
                        if not PathGeom.isRoughly(p1.z, minZ):
                            # the edges covered by the helix not handled again,
                            # unless reached the bottom height
                            i = j

                else:
                    outedges.append(edge)
            else:
                outedges.append(edge)
            i = i + 1
        return outedges

    def checkIgnoreAbove(self, edge):
        if self.ignoreAboveEnabled:
            p0 = edge.Vertexes[0].Point
            p1 = edge.Vertexes[1].Point
            if p0.z > self.ignoreAbove and (p1.z > self.ignoreAbove or PathGeom.isRoughly(p1.z, self.ignoreAbove.Value)):
                PathLog.debug("Whole plunge move above 'ignoreAbove', ignoring")
                return (edge, True)
            elif p0.z > self.ignoreAbove and not PathGeom.isRoughly(p0.z, self.ignoreAbove.Value):
                PathLog.debug("Plunge move partially above 'ignoreAbove', splitting into two")
                newPoint = FreeCAD.Base.Vector(p0.x, p0.y, self.ignoreAbove)
                return (Part.makeLine(p0, newPoint), False)
            else:
                return None, False
        else:
            return None, False

    def createHelix(self, rampedges, startPoint, endPoint):
        outedges = []
        ramplen = 0
        for redge in rampedges:
            ramplen = ramplen + redge.Length
        rampheight = abs(endPoint.z - startPoint.z)
        rampangle_rad = math.atan(ramplen / rampheight)
        curPoint = startPoint
        for i, redge in enumerate(rampedges):
            if i < len(rampedges) - 1:
                deltaZ = redge.Length / math.tan(rampangle_rad)
                newPoint = FreeCAD.Base.Vector(redge.valueAt(redge.LastParameter).x, redge.valueAt(redge.LastParameter).y, curPoint.z - deltaZ)
                outedges.append(self.createRampEdge(redge, curPoint, newPoint))
                curPoint = newPoint
            else:
                # on the last edge, force it to end to the endPoint
                # this should happen automatically, but this avoids any rounding error
                outedges.append(self.createRampEdge(redge, curPoint, endPoint))
        return outedges

    def createRampEdge(self, originalEdge, startPoint, endPoint):
        # PathLog.debug("Create edge from [{},{},{}] to [{},{},{}]".format(startPoint.x,startPoint.y, startPoint.z, endPoint.x, endPoint.y, endPoint.z))
        if type(originalEdge.Curve) == Part.Line or type(originalEdge.Curve) == Part.LineSegment:
            return Part.makeLine(startPoint, endPoint)
        elif type(originalEdge.Curve) == Part.Circle:
            arcMid = originalEdge.valueAt((originalEdge.FirstParameter + originalEdge.LastParameter) / 2)
            arcMid.z = (startPoint.z + endPoint.z) / 2
            return Part.Arc(startPoint, arcMid, endPoint).toShape()
        else:
            PathLog.error("Edge should not be helix")

    def getreversed(self, edges):
        """
        Reverses the edge array and the direction of each edge
        """
        outedges = []
        for edge in reversed(edges):
            # reverse the start and end points
            startPoint = edge.valueAt(edge.LastParameter)
            endPoint = edge.valueAt(edge.FirstParameter)
            if type(edge.Curve) == Part.Line or type(edge.Curve) == Part.LineSegment:
                outedges.append(Part.makeLine(startPoint, endPoint))
            elif type(edge.Curve) == Part.Circle:
                arcMid = edge.valueAt((edge.FirstParameter + edge.LastParameter) / 2)
                outedges.append(Part.Arc(startPoint, arcMid, endPoint).toShape())
            else:
                PathLog.error("Edge should not be helix")
        return outedges

    def findMinZ(self, edges):
        minZ = 99999999999
        for edge in edges[1:]:
            for v in edge.Vertexes:
                if v.Point.z < minZ:
                    minZ = v.Point.z
        return minZ

    def getSplitPoint(self, edge, remaining):
        if type(edge.Curve) == Part.Line or type(edge.Curve) == Part.LineSegment:
            return edge.valueAt(remaining)
        elif type(edge.Curve) == Part.Circle:
            param = remaining / edge.Curve.Radius
            return edge.valueAt(param)

    def createRampMethod1(self, rampedges, p0, projectionlen, rampangle):
        """
        This method generates ramp with following pattern:
        1. Start from the original startpoint of the plunge
        2. Ramp down along the path that comes after the plunge
        3. When reaching the Z level of the original plunge, return back to the beginning
           by going the path backwards until the original plunge end point is reached
        4. Continue with the original path

        This method causes many unnecessary moves with tool down.
        """
        outedges = []
        rampremaining = projectionlen
        curPoint = p0  # start from the upper point of plunge
        done = False
        goingForward = True
        while not done:
            for i, redge in enumerate(rampedges):
                if redge.Length >= rampremaining:
                    # will reach end of ramp within this edge, needs to be split
                    p1 = self.getSplitPoint(redge, rampremaining)
                    splitEdge = PathGeom.splitEdgeAt(redge, p1)
                    PathLog.debug("Ramp remaining: {}".format(rampremaining))
                    PathLog.debug("Got split edge (index: {}) (total len: {}) with lengths: {}, {}".format(i, redge.Length, splitEdge[0].Length, splitEdge[1].Length))
                    # ramp ends to the last point of first edge
                    p1 = splitEdge[0].valueAt(splitEdge[0].LastParameter)
                    outedges.append(self.createRampEdge(splitEdge[0], curPoint, p1))
                    # now we have reached the end of the ramp. Go back to plunge position with constant Z
                    # start that by going to the beginning of this splitEdge
                    if goingForward:
                        outedges.append(self.createRampEdge(splitEdge[0], p1, redge.valueAt(redge.FirstParameter)))
                    else:
                        # if we were reversing, we continue to the same direction as the ramp
                        outedges.append(self.createRampEdge(splitEdge[0], p1, redge.valueAt(redge.LastParameter)))
                    done = True
                    break
                else:
                    deltaZ = redge.Length / math.tan(math.radians(rampangle))
                    newPoint = FreeCAD.Base.Vector(redge.valueAt(redge.LastParameter).x, redge.valueAt(redge.LastParameter).y, curPoint.z - deltaZ)
                    outedges.append(self.createRampEdge(redge, curPoint, newPoint))
                    curPoint = newPoint
                    rampremaining = rampremaining - redge.Length

            if not done:
                # we did not reach the end of the ramp going this direction, lets reverse.
                rampedges = self.getreversed(rampedges)
                PathLog.debug("Reversing")
                if goingForward:
                    goingForward = False
                else:
                    goingForward = True
        # now we need to return to original position.
        if goingForward:
            # if the ramp was going forward, the return edges are the edges we already covered in ramping,
            # except the last one, which was already covered inside for loop. Direction needs to be reversed also
            returnedges = self.getreversed(rampedges[:i])
        else:
            # if the ramp was already reversing, the edges needed for return are the ones
            # which were not covered in ramp
            returnedges = rampedges[(i + 1):]

        # add the return edges:
        outedges.extend(returnedges)

        return outedges

    def createRampMethod3(self, rampedges, p0, projectionlen, rampangle):
        """
        This method generates ramp with following pattern:
        1. Start from the original startpoint of the plunge
        2. Ramp down along the path that comes after the plunge until
           traveled half of the Z distance
        3. Change direction and ramp backwards to the original plunge end point
        4. Continue with the original path

        This method causes many unnecessary moves with tool down.
        """
        outedges = []
        rampremaining = projectionlen
        curPoint = p0  # start from the upper point of plunge
        done = False

        while not done:
            for i, redge in enumerate(rampedges):
                if redge.Length >= rampremaining:
                    # will reach end of ramp within this edge, needs to be split
                    p1 = self.getSplitPoint(redge, rampremaining)
                    splitEdge = PathGeom.splitEdgeAt(redge, p1)
                    PathLog.debug("Got split edge (index: {}) with lengths: {}, {}".format(i, splitEdge[0].Length, splitEdge[1].Length))
                    # ramp ends to the last point of first edge
                    p1 = splitEdge[0].valueAt(splitEdge[0].LastParameter)
                    deltaZ = splitEdge[0].Length / math.tan(math.radians(rampangle))
                    p1.z = curPoint.z - deltaZ
                    outedges.append(self.createRampEdge(splitEdge[0], curPoint, p1))
                    curPoint.z = p1.z - deltaZ
                    # now we have reached the end of the ramp. Reverse direction of ramp
                    # start that by going back to the beginning of this splitEdge
                    outedges.append(self.createRampEdge(splitEdge[0], p1, curPoint))

                    done = True
                    break
                elif i == len(rampedges) - 1:
                    # last ramp element but still did not reach the full length?
                    # Probably a rounding issue on floats.
                    p1 = redge.valueAt(redge.LastParameter)
                    deltaZ = redge.Length / math.tan(math.radians(rampangle))
                    p1.z = curPoint.z - deltaZ
                    outedges.append(self.createRampEdge(redge, curPoint, p1))
                    # and go back that edge
                    newPoint = FreeCAD.Base.Vector(redge.valueAt(redge.FirstParameter).x, redge.valueAt(redge.FirstParameter).y, p1.z - deltaZ)
                    outedges.append(self.createRampEdge(redge, p1, newPoint))
                    curPoint = newPoint
                    done = True
                else:
                    deltaZ = redge.Length / math.tan(math.radians(rampangle))
                    newPoint = FreeCAD.Base.Vector(redge.valueAt(redge.LastParameter).x, redge.valueAt(redge.LastParameter).y, curPoint.z - deltaZ)
                    outedges.append(self.createRampEdge(redge, curPoint, newPoint))
                    curPoint = newPoint
                    rampremaining = rampremaining - redge.Length

        returnedges = self.getreversed(rampedges[:i])

        # ramp backwards to the plunge position
        for i, redge in enumerate(returnedges):
            deltaZ = redge.Length / math.tan(math.radians(rampangle))
            newPoint = FreeCAD.Base.Vector(redge.valueAt(redge.LastParameter).x, redge.valueAt(redge.LastParameter).y, curPoint.z - deltaZ)
            if i == len(rampedges) - 1:
                # make sure that the last point of the ramps ends to the original position
                newPoint = redge.valueAt(redge.LastParameter)
            outedges.append(self.createRampEdge(redge, curPoint, newPoint))
            curPoint = newPoint

        return outedges

    def createRampMethod2(self, rampedges, p0, projectionlen, rampangle):
        """
        This method generates ramp with following pattern:
        1. Start from the original startpoint of the plunge
        2. Calculate the distance on the path which is needed to implement the ramp
           and travel that distance while maintaining start depth
        3. Start ramping while travelling the original path backwards until reaching the
           original plunge end point
        4. Continue with the original path
        """
        outedges = []
        rampremaining = projectionlen
        curPoint = p0  # start from the upper point of plunge
        if PathGeom.pointsCoincide(PathGeom.xy(p0), PathGeom.xy(rampedges[-1].valueAt(rampedges[-1].LastParameter))):
            PathLog.debug("The ramp forms a closed wire, needless to move on original Z height")
        else:
            for i, redge in enumerate(rampedges):
                if redge.Length >= rampremaining:
                    # this edge needs to be split
                    p1 = self.getSplitPoint(redge, rampremaining)
                    splitEdge = PathGeom.splitEdgeAt(redge, p1)
                    PathLog.debug("Got split edges with lengths: {}, {}".format(splitEdge[0].Length, splitEdge[1].Length))
                    # ramp starts at the last point of first edge
                    p1 = splitEdge[0].valueAt(splitEdge[0].LastParameter)
                    p1.z = p0.z
                    outedges.append(self.createRampEdge(splitEdge[0], curPoint, p1))
                    # now we have reached the beginning of the ramp.
                    # start that by going to the beginning of this splitEdge
                    deltaZ = splitEdge[0].Length / math.tan(math.radians(rampangle))
                    newPoint = FreeCAD.Base.Vector(splitEdge[0].valueAt(splitEdge[0].FirstParameter).x, splitEdge[0].valueAt(splitEdge[0].FirstParameter).y, p1.z - deltaZ)
                    outedges.append(self.createRampEdge(splitEdge[0], p1, newPoint))
                    curPoint = newPoint
                elif i == len(rampedges) - 1:
                    # last ramp element but still did not reach the full length?
                    # Probably a rounding issue on floats.
                    # Lets start the ramp anyway
                    p1 = redge.valueAt(redge.LastParameter)
                    p1.z = p0.z
                    outedges.append(self.createRampEdge(redge, curPoint, p1))
                    # and go back that edge
                    deltaZ = redge.Length / math.tan(math.radians(rampangle))
                    newPoint = FreeCAD.Base.Vector(redge.valueAt(redge.FirstParameter).x, redge.valueAt(redge.FirstParameter).y, p1.z - deltaZ)
                    outedges.append(self.createRampEdge(redge, p1, newPoint))
                    curPoint = newPoint

                else:
                    # we are travelling on start depth
                    newPoint = FreeCAD.Base.Vector(redge.valueAt(redge.LastParameter).x, redge.valueAt(redge.LastParameter).y, p0.z)
                    outedges.append(self.createRampEdge(redge, curPoint, newPoint))
                    curPoint = newPoint
                    rampremaining = rampremaining - redge.Length

            # the last edge got handled previously
            rampedges.pop()
        # ramp backwards to the plunge position
        for i, redge in enumerate(reversed(rampedges)):
            deltaZ = redge.Length / math.tan(math.radians(rampangle))
            newPoint = FreeCAD.Base.Vector(redge.valueAt(redge.FirstParameter).x, redge.valueAt(redge.FirstParameter).y, curPoint.z - deltaZ)
            if i == len(rampedges) - 1:
                # make sure that the last point of the ramps ends to the original position
                newPoint = redge.valueAt(redge.FirstParameter)
            outedges.append(self.createRampEdge(redge, curPoint, newPoint))
            curPoint = newPoint

        return outedges

    def createCommands(self, obj, edges):
        commands = []
        for edge in edges:
            israpid = False
            for redge in self.rapids:
                if PathGeom.edgesMatch(edge, redge):
                    israpid = True
            if israpid:
                v = edge.valueAt(edge.LastParameter)
                commands.append(Path.Command('G0', {'X': v.x, 'Y': v.y, 'Z': v.z}))
            else:
                commands.extend(PathGeom.cmdsForEdge(edge))

        lastCmd = Path.Command('G0', {'X': 0.0, 'Y': 0.0, 'Z': 0.0})

        outCommands = []

        tc = PathDressup.toolController(obj.Base)

        horizFeed = tc.HorizFeed.Value
        vertFeed = tc.VertFeed.Value
        if obj.RampFeedRate == "Horizontal Feed Rate":
            rampFeed = tc.HorizFeed.Value
        elif obj.RampFeedRate == "Vertical Feed Rate":
            rampFeed = tc.VertFeed.Value
        else:
            rampFeed = obj.CustomFeedRate.Value
        horizRapid = tc.HorizRapid.Value
        vertRapid = tc.VertRapid.Value

        for cmd in commands:
            params = cmd.Parameters
            zVal = params.get('Z', None)
            zVal2 = lastCmd.Parameters.get('Z', None)

            xVal = params.get('X', None)
            xVal2 = lastCmd.Parameters.get('X', None)

            yVal2 = lastCmd.Parameters.get('Y', None)
            yVal = params.get('Y', None)

            zVal = zVal and round(zVal, 8)
            zVal2 = zVal2 and round(zVal2, 8)

            if cmd.Name in ['G1', 'G2', 'G3', 'G01', 'G02', 'G03']:
                if zVal is not None and zVal2 != zVal:
                    if PathGeom.isRoughly(xVal, xVal2) and PathGeom.isRoughly(yVal, yVal2):
                        # this is a straight plunge
                        params['F'] = vertFeed
                    else:
                        # this is a ramp
                        params['F'] = rampFeed
                else:
                    params['F'] = horizFeed
                lastCmd = cmd

            elif cmd.Name in ['G0', 'G00']:
                if zVal is not None and zVal2 != zVal:
                    params['F'] = vertRapid
                else:
                    params['F'] = horizRapid
                lastCmd = cmd

            outCommands.append(Path.Command(cmd.Name, params))

        return Path.Path(outCommands)


class ViewProviderDressup:

    def __init__(self, vobj):
        vobj.Proxy = self

    def attach(self, vobj):
        self.obj = vobj.Object

    def claimChildren(self):
        if hasattr(self.obj.Base, "InList"):
            for i in self.obj.Base.InList:
                if hasattr(i, "Group"):
                    group = i.Group
                    for g in group:
                        if g.Name == self.obj.Base.Name:
                            group.remove(g)
                    i.Group = group
                    print(i.Group)
        # FreeCADGui.ActiveDocument.getObject(obj.Base.Name).Visibility = False
        return [self.obj.Base]

    def onDelete(self, arg1=None, arg2=None):
        PathLog.debug("Deleting Dressup")
        '''this makes sure that the base operation is added back to the project and visible'''
        FreeCADGui.ActiveDocument.getObject(arg1.Object.Base.Name).Visibility = True
        job = PathUtils.findParentJob(self.obj)
        job.Proxy.addOperation(arg1.Object.Base, arg1.Object)
        arg1.Object.Base = None
        return True

    def __getstate__(self):
        return None

    def __setstate__(self, state):
        return None


class CommandPathDressupRampEntry:

    def GetResources(self):
        return {'Pixmap': 'Path-Dressup',
                'MenuText': QtCore.QT_TRANSLATE_NOOP("Path_DressupRampEntry", "RampEntry Dress-up"),
                'ToolTip': QtCore.QT_TRANSLATE_NOOP("Path_DressupRampEntry", "Creates a Ramp Entry Dress-up object from a selected path")}

    def IsActive(self):
        op = PathDressup.selection()
        if op:
            return not PathDressup.hasEntryMethod(op)
        return False

    def Activated(self):

        # check that the selection contains exactly what we want
        selection = FreeCADGui.Selection.getSelection()
        if len(selection) != 1:
            PathLog.error(translate("Please select one path object") + "\n")
            return
        baseObject = selection[0]
        if not baseObject.isDerivedFrom("Path::Feature"):
            PathLog.error(translate("The selected object is not a path") + "\n")
            return
        if baseObject.isDerivedFrom("Path::FeatureCompoundPython"):
            PathLog.error(translate("Please select a Profile object"))
            return

        # everything ok!
        FreeCAD.ActiveDocument.openTransaction(translate("Create RampEntry Dress-up"))
        FreeCADGui.addModule("PathScripts.PathDressupRampEntry")
        FreeCADGui.addModule("PathScripts.PathUtils")
        FreeCADGui.doCommand('obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", "RampEntryDressup")')
        FreeCADGui.doCommand('dbo = PathScripts.PathDressupRampEntry.ObjectDressup(obj)')
        FreeCADGui.doCommand('base = FreeCAD.ActiveDocument.' + selection[0].Name)
        FreeCADGui.doCommand('job = PathScripts.PathUtils.findParentJob(base)')
        FreeCADGui.doCommand('obj.Base = base')
        FreeCADGui.doCommand('job.Proxy.addOperation(obj, base)')
        FreeCADGui.doCommand('PathScripts.PathDressupRampEntry.ViewProviderDressup(obj.ViewObject)')
        FreeCADGui.doCommand('Gui.ActiveDocument.getObject(base.Name).Visibility = False')
        FreeCADGui.doCommand('dbo.setup(obj)')
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand('Path_DressupRampEntry', CommandPathDressupRampEntry())

PathLog.notice("Loading Path_DressupRampEntry... done\n")



# dom3ds.py

"""Slice and dice 3DS files.

Provides for reading, writing, and manipulating 3DS files.  It's
called dom3ds because it's reminiscent of XML-DOM: it converts the 3DS
file into a hierarchy of objects, in much the same way XML-DOM
converts an XML file into a hierarchy of objects called the Document
Object Model.  The dom3ds module creates an object for each chunk in
the 3DS file, which can be accessed hierarchially as attributes.

For example, once a 3DS file is loaded, you could access the smoothing
data of the second object like this:

    dom.mdata.objects[2].ntri.faces.smoothing.array

"""

import os, sys, struct

import numpy

# xrange is not available in python3
if sys.version_info.major >= 3:
    xrange = range

# Exceptions

class File3dsFormatError(Exception):
    "Exception in the 3DS file format"
    def __init__(self,msg,tight=False):
        self.msg = msg
        self.tight = tight
    def __str__(self):
        return repr(self.msg)

class FBufError(Exception):
    "Exception when reading from FileLikeBuffers"


# Numeric/string conversion functions, where the string uses
# little-endian regardless of host byte order

def string_to_array(s,typ):
    """Convert data block from a 3DS file to a numpy array."""
    a = numpy.fromstring(s,typ)
    if sys.byteorder == 'big':
        a.byteswap(True)
    return a

def array_to_string(a):
    """Convert a numpy array to data block for a 3DS file."""
    if sys.byteorder == 'big':
        a = a.byteswap()
    return a.tostring()

def array_to_string_destructive(a):
    """Destructively convert a numpy array to data block for a 3DS file."""
    if sys.byteorder == 'big':
        a.byteswap(True)
    return a.tostring()


# Utility classes

class FileLikeBuffer(object):
    """Treat a range of a string as a quasi-file-like object.

    This is a quasi-file like object that represents a certain
    range of a string.  It behaves kind of like a file object,
    but has some specialized methods.

    """

    def __init__(self,s,start=0,end=None):
        self.buf = s
        self.start = start
        if end is None:
            self.end = len(s)
        else:
            if end > len(s):
                raise FBufError("End of fbuf object set past end of buffer")
            self.end = end
        assert 0 <= start <= self.end, (start,self.end)
        self.index = start
    def tell(self):
        return self.index
    def seek(self,pos):
        self.index = pos
        if self.index < 0 or self.index > self.end:
            raise FBufError("Invalid position in fbuf")    
    def advance(self,n):
        self.index += n
        if self.index > self.end:
            raise FBufError("End of fbuf reached")
    def read(self,length):
        r = self.buf[self.index:self.index+length]
        self.advance(length)
        return r
    def read_fbuf(self,length):
        r = FileLikeBuffer(self.buf,self.index,self.index+length)
        self.advance(length)
        return r
    def read_to_nul(self):
        try:
            i = self.buf.index('\0',self.index,self.end)
        except IndexError:
            raise FBufError("End of fbuf reached in string")
        r = self.buf[self.index:i]
        self.index = i+1
        return r
    def read_rest(self):
        r = self.buf[self.index:self.end]
        self.index = self.end
        return r
    def room_for_chunks(self):
        return self.index+6 <= self.end


# Utility functions

def _decls_to_list(decls):
    decllist = [ tuple(x.strip().split()) for x in decls.split(',') ]
    return decllist

def _decls_to_vars(decls):
    if not decls:
        return ()
    vars = [ x.strip().split(None,1)[1] for x in decls.split(',') ]
    return vars


# Chunk metaclasses

class ChunkMetaclass(type):
    """Metaclass of chunk types.

    This class takes some static data and defines how to read in
    the chunks, and preprocesses it.  It also creates a dictionary
    of chunk classes keyed by tag.

    """

    chunk_taghash = {}
    chunk_labelhash = {}

    def __new__(metatype,name,bases,clsdict):
        vars = []
        vars.extend(_decls_to_vars(clsdict.get('struct','')))
        vars.extend(_decls_to_vars(clsdict.get('single','')))
        vars.extend(_decls_to_vars(clsdict.get('multiple','')))
        if clsdict.get('freechunks',False):
            vars.append('subchunks')
        clsdict.setdefault('__slots__',[]).extend(vars)

        return type.__new__(metatype,name,bases,clsdict)

    def __init__(self,name,bases,clsdict):
        if hasattr(self,'struct'):
            self.struct_fields = _decls_to_list(self.struct)
        else:
            self.struct_fields = []

        self.single_types = {}
        self.single_order = []
        if hasattr(self,'single'):
            attrs = {}
            for typ,attr in _decls_to_list(self.single):
                self.single_types[typ] = attr
                if attr not in attrs:
                    attrs[attr] = 1
                    self.single_order.append(attr)

        self.multiple_types = {}
        self.multiple_order = []
        if hasattr(self,'multiple'):
            attrs = {}
            for typ,attr in _decls_to_list(self.multiple):
                self.multiple_types[typ] = attr
                if attr not in attrs:
                    attrs[attr] = 1
                    self.multiple_order.append(attr)

        if hasattr(self,'keyframe'):
            self.keyframe_fields = _decls_to_list(self.keyframe)
            slots = ['frameno','flags','tension',
                     'continuity','bias','ease_to','ease_from']
            for typ,attr in self.keyframe_fields:
                slots.append(attr)
            kfdict = { '__slots__': slots }
            kfbases = (object,)
            kfname = "%s.Key" % name
            self.Key = type(kfname,kfbases,kfdict)
        else:
            self.keyframe_fields = []

        if hasattr(self,'tag'):
            self.chunk_taghash[self.tag] = self
            self.chunk_labelhash[name] = self
            if not hasattr(self,'label'):
                self.label = name

        type.__init__(self,name,bases,clsdict)



class UndefinedChunkMetaclass(ChunkMetaclass):
    """Metaclass of undefined chunks.

    Automatically inserts a data attribute into __slots__ to hold
    the chunk data.

    """

    def __new__(metatype,name,bases,clsdict):
        if '__slots__' in clsdict:
            if 'data' not in clsdict['__slots__']:
                clsdict['__slots__'].append('data')
        else:
            clsdict['__slots__'] = ['data']
        return ChunkMetaclass.__new__(metatype,name,bases,clsdict)



# Base class of chunks.

class ChunkBase(object):
    """Base class of chunks.

    The methods of this class are mostly for input/output.  When
    manipulating a 3ds model, one would access subchunks as
    attributes.

    """

    __metaclass__ = ChunkMetaclass

    # Defaults for some parameters

    freechunks = True
    swallow = False

    # Define constructor to accept keyword arguments, as a convenience
    # Init chunk arguments to nothing

    def __init__(self,**kwargs):
        """Construct a chunk.

        This is an abstract base constructor; normally you'd
        construct a subclass of this.  As a convenience, you
        can set attributes via keyword arguments when creating
        a chunk.

        """

        if self.freechunks:
            self.subchunks = []
        for v in self.single_order:
            setattr(self,v,None)
        for v in self.multiple_order:
            setattr(self,v,[])
        for attr,value in kwargs.items():
            setattr(self,attr,value)

    # -------- SECTION 1: INITIALIZATION FROM CHUNK DATA --------

    # These get a specific object from the fbuf

    @staticmethod
    def get_byte(fbuf,flags):
        return ord(fbuf.read(1))

    @staticmethod
    def get_short(fbuf,flags):
        return struct.unpack("<H",fbuf.read(2))[0]

    @staticmethod
    def get_long(fbuf,flags):
        return struct.unpack("<L",fbuf.read(4))[0]

    @staticmethod
    def get_float(fbuf,flags):
        return struct.unpack("<f",fbuf.read(4))[0]

    @staticmethod
    def get_string(fbuf,flags):
        return fbuf.read_to_nul()

    @staticmethod
    def get_chunk(fbuf,flags):
        tag,length = struct.unpack("<HL",fbuf.read(6))
        pos = fbuf.tell()
        cls = ChunkMetaclass.chunk_taghash.get(tag)
        try:
            if cls is not None:
                ch = cls()
            else:
                ch = UnknownChunk(tag)
            ch.read(fbuf.read_fbuf(length-6),flags)
        except (File3dsFormatError, FBufError) as fe:
            if flags['recover']:
                if cls is not None:
                    label = cls.label
                else:
                    label = "UNKNOWN"
                ch = ErrorChunk(tag,label,str(fe))
                fbuf.seek(pos)
                ch.read(fbuf.read_fbuf(length-6),flags)
            else:
                raise
        return ch

    # These read sections of the chunk

    def read_struct(self,fbuf,flags):
        for typ,attr in self.struct_fields:
            setattr(self,attr,getattr(self,"get_"+typ)(fbuf,flags))

    def read_array(self,fbuf,flags):
        # Expect subclass to override this
        pass

    def read_chunks(self,fbuf,flags):
        if self.swallow:
            return

        while fbuf.room_for_chunks():
            ch = self.get_chunk(fbuf,flags)
            label = ch.label
            if label == "ERROR":
                label = ch.intended_label
                if label is None:
                    label = "ERROR"
            if label in self.single_types:
                attr = self.single_types[label]
                if flags['tight'] and getattr(self,attr) is not None:
                    raise File3dsFormatError(
                        "Chunk %s appeared twice" % label,True)
                setattr(self,attr,ch)
            elif label in self.multiple_types:
                attr = self.multiple_types[label]
                getattr(self,attr).append(ch)
            elif label == 'DUMMY':
                pass
            elif self.freechunks:
                self.subchunks.append(ch)
            elif flags['tight']:
                raise File3dsFormatError(
                    "Illegal context for chunk %s" % label,True)

    # Initialize self by scanning a chunk

    def read(self,fbuf,flags):
        self.read_struct(fbuf,flags)
        self.read_array(fbuf,flags)
        self.read_chunks(fbuf,flags)

    # -------- SECTION 2: DUMPING TO OUTPUT --------

    # Dump individual attributes

    def out_attr(self,attr,flo,indent,flags):
        flo.write("%s%s = %r\n" % (indent,attr,getattr(self,attr)))

    def out_chunk(self,attr,flo,indent,flags):
        g = getattr(self,attr)
        if g is not None:
            flo.write("%s%s = " % (indent,attr))
            g.dump(flo,indent,flags)

    def out_chunk_index(self,attr,i,flo,indent,flags):
        g = getattr(self,attr)[i]
        if g is not None:
            flo.write("%s%s[%d] = " % (indent,attr,i))
            g.dump(flo,indent,flags)


    # Dump sections of the file

    def dump_header(self,flo,flags):
        flo.write("%s (0x%04X)\n" % (self.label,self.tag))

    def dump_struct(self,flo,indent,flags):
        for typ,attr in self.struct_fields:
            self.out_attr(attr,flo,indent,flags)

    def dump_array(self,flo,indent,flags):
        # Expect subclass to override this
        pass

    def dump_chunks(self,flo,indent,flags):
        if self.swallow:
            return
        for attr in self.single_order:
            self.out_chunk(attr,flo,indent,flags)
        for attr in self.multiple_order:
            for i in xrange(len(getattr(self,attr))):
                self.out_chunk_index(attr,i,flo,indent,flags)
        if self.freechunks:
            attr = "subchunks"
            for i in xrange(len(self.subchunks)):
                self.out_chunk_index(attr,i,flo,indent,flags)

    # Dump self to stream flo at a certain indentation

    def dump(self,flo,indent,flags):
        indent += '    '
        self.dump_header(flo,flags)
        self.dump_struct(flo,indent,flags)
        self.dump_array(flo,indent,flags)
        self.dump_chunks(flo,indent,flags)


    # -------- SECTION #3: WRITING TO OUTPUT --------

    # These set a specific object from the fbuf

    @staticmethod
    def set_byte(value):
        return chr(value)

    @staticmethod
    def set_short(value):
        return struct.pack("<H",value)

    @staticmethod
    def set_long(value):
        return struct.pack("<L",value)

    @staticmethod
    def set_float(value):
        return struct.pack("<f",value)

    @staticmethod
    def set_string(value):
        return value + "\0"

    @staticmethod
    def set_chunk(value):
        s = value.write()
        length = len(s) + 6
        return struct.pack("<HL",value.tag,length) + s

    # These write sections of the chunk

    def write_struct(self):
        s = []
        for typ,attr in self.struct_fields:
            s.append(getattr(self,"set_"+typ)(getattr(self,attr)))
        return ''.join(s)

    def write_array(self):
        # Expect subclass to override this
        return ''

    def write_chunks(self):
        if self.swallow:
            return ''
        s = []
        for attr in self.single_order:
            v = getattr(self,attr)
            if v is not None:
                s.append(self.set_chunk(v))
        for attr in self.multiple_order:
            for v in getattr(self,attr):
                s.append(self.set_chunk(v))
        if self.freechunks:
            for v in self.subchunks:
                s.append(self.set_chunk(v))
        return ''.join(s)

    # Write out contents (but not header) to chunk

    def write(self):
        s = []
        s.append(self.write_struct())
        s.append(self.write_array())
        s.append(self.write_chunks())
        return ''.join(s)


    # -------- SECTION #4: OUTPUT DOCUMENTATION --------

    def document_html(cls,flo,done):
        if cls.label in done:
            return
        flo.write('<dt>\n<a name="%s"></a><b><tt>%s (0x%04X)</tt></b></dt>\n'
                  % (cls.label,cls.label,cls.tag))
        flo.write('<dd>\n')

        i = 0

        for typ,attr in cls.struct_fields:
            flo.write('<tt>%s %s</tt><br>\n' % (typ,attr))
            i += 1

        tagorder = []

        d = {}
        for typ,attr in cls.single_types.items():
            if attr not in d:
                d[attr] = []
            d[attr].append(typ)
        for attr in cls.single_order:
            typs = d[attr]
            for typ in typs:
                tagorder.append(typ)
                flo.write('<tt><a href="#%s">%s</a> %s</tt><br>\n'
                          % (typ,typ,attr))
                i += 1

        d = {}
        for typ,attr in cls.multiple_types.items():
            if attr not in d:
                d[attr] = []
            d[attr].append(typ)
        for attr in cls.multiple_order:
            typs = d[attr]
            for typ in typs:
                tagorder.append(typ)
                flo.write('<tt><a href="#%s">%s[]</a> %s</tt><br>\n'
                          % (typ,typ,attr))
                i += 1

        if not i:
            flo.write('No fields<br>\n')

        flo.write('&nbsp;</dd>\n')
        done[cls.label] = 1

        for typ in tagorder:
            type(cls).chunk_labelhash[typ].document_html(flo,done)

    document_html = classmethod(document_html)


# Chunks we don't know anything about, maybe not even the tag

class UndefinedChunk(ChunkBase):
    __metaclass__ = UndefinedChunkMetaclass
    def read(self,fbuf,flags):
        self.data = fbuf.read_rest()
    def dump(self,flo,indent,flags):
        self.dump_header(flo,flags)
    def write(self):
        return self.data


class UnknownChunk(UndefinedChunk):
    __slots__ = ['tag']
    label = "UNKNOWN"
    def __init__(self,tag):
        self.tag = tag


class ErrorChunk(UndefinedChunk):
    __slots__ = ['intended_tag','intended_label','error_msg']
    label = "ERROR"
    tag = 0xEEEE
    def __init__(self,intended_tag=None,intended_label=None,error_msg=None):
        self.intended_tag = intended_tag
        self.intended_label = intended_label
        self.error_msg = error_msg
    def dump(self,flo,indent,flags):
        super(ErrorChunk,self).dump(flo,indent,flags)
        if self.intended_tag is not None:
            flo.write("%s    intended_tag = 0x%04X\n"
                      % (indent,self.intended_tag))
        if self.intended_label is not None:
            self.out_attr("intended_label",flo,indent+"    ",flags)
        if self.error_msg is not None:
            self.out_attr("error_msg",flo,indent+"    ",flags)
    def write(self):
        raise TypeError("Attempt to write an Error Chunk")


# Chunks with common sets of attributes

class OneColorChunk(ChunkBase):
    single = ("COLOR_F color, COLOR_24 color,"
              "LIN_COLOR_F lincolor, LIN_COLOR_24 lincolor")

class OnePercentageChunk(ChunkBase):
    single = ("INT_PERCENTAGE pct, FLOAT_PERCENTAGE pct")

class OneShortValueChunk(ChunkBase):
    struct = "short value"

class OneFloatValueChunk(ChunkBase):
    struct = "float value"

class Color24Chunk(ChunkBase):
    struct = "byte red, byte green, byte blue"

class TextureChunk(ChunkBase):
    single = ("INT_PERCENTAGE pct,"
              "MAT_MAPNAME filename,"
              "MAT_MAP_TILING tiling,"
              "MAT_MAP_TEXBLUR blur,"
              "MAT_MAP_USCALE uscale,"
              "MAT_MAP_VSCALE vscale,"
              "MAT_MAP_UOFFSET uoffset,"
              "MAT_MAP_VOFFSET voffset,"
              "MAT_MAP_ANG angle,"
              "MAT_MAP_COL1 color1,"
              "MAT_MAP_COL2 color2,"
              "MAT_MAP_RCOL rtint,"
              "MAT_MAP_GCOL gtint,"
              "MAT_MAP_BCOL btint")

class TextureMaskChunk(ChunkBase):
    single = ("INT_PERCENTAGE pct,"
              "MAT_MAPNAME filename,"
              "MAT_MAP_TILING tiling,"
              "MAT_MAP_TEXBLUR blur,"
              "MAT_MAP_USCALE uscale,"
              "MAT_MAP_VSCALE vscale,"
              "MAT_MAP_UOFFSET uoffset,"
              "MAT_MAP_VOFFSET voffset,"
              "MAT_MAP_ANG angle")

class ArrayChunk(ChunkBase):
    __slots__ = ['array']
    def dump_array(self,flo,indent,flags):
        flo.write("%s%s = <array data %s>\n"
                  % (indent,"array",repr(self.array.shape)))

class MatrixChunk(ArrayChunk):
    def read_array(self,fbuf,flags):
        size = 48
        a = string_to_array(fbuf.read(size),numpy.float32)
        a = numpy.reshape(a,(4,3))
        m = numpy.zeros((4,4),numpy.float32)
        m[:,0:3] = a
        m[3,3] = 1.0
        self.array = numpy.array(numpy.transpose(m))
    def dump_array(self,flo,indent,flags):
        super(MatrixChunk,self).dump_array(flo,indent,flags)
        if flags['arraylines'] == 0:
            return
        for i in xrange(4):
            flo.write("%s    %12.4g%12.4g%12.4g%12.4g\n"
                      % (indent,self.array[i,0],self.array[i,1],
                         self.array[i,2],self.array[i,3]))
    def write_array(self):
        a = self.array[0:3,:]
        a = numpy.array(numpy.transpose(a)).astype(numpy.float32)
        return array_to_string_destructive(a)

class TrackChunk(ChunkBase):
    __slots__ = ['keys']
    struct = "short flags, long unused1, long unused2, long nkeys"
    def read_array(self,fbuf,flags):
        self.keys = []
        for i in xrange(self.nkeys):
            kf = self.Key()
            kf.frameno = self.get_long(fbuf,flags)
            kf.flags = self.get_short(fbuf,flags)
            if kf.flags & 1:
                kf.tension = self.get_float(fbuf,flags)
            if kf.flags & 2:
                kf.continuity = self.get_float(fbuf,flags)
            if kf.flags & 4:
                kf.bias = self.get_float(fbuf,flags)
            if kf.flags & 8:
                kf.ease_to = self.get_float(fbuf,flags)
            if kf.flags & 16:
                kf.ease_from = self.get_float(fbuf,flags)
            for typ,attr in self.keyframe_fields:
                setattr(kf,attr,getattr(self,"get_"+typ)(fbuf,flags))
            self.keys.append(kf)
    def write_array(self):
        s = []
        for kf in self.keys:
            s.append(self.set_long(kf.frameno))
            s.append(self.set_short(kf.flags))
            if kf.flags & 1:
                s.append(self.set_float(kf.tension))
            if kf.flags & 2:
                s.append(self.set_float(kf.continuity))
            if kf.flags & 4:
                s.append(self.set_float(kf.bias))
            if kf.flags & 8:
                s.append(self.set_float(kf.ease_to))
            if kf.flags & 16:
                s.append(self.set_float(kf.ease_from))
            for typ,attr in self.keyframe_fields:
                s.append(getattr(self,"set_"+typ)(getattr(kf,attr)))
        return "".join(s)
    def dump_array(self,flo,indent,flags):
        super(TrackChunk,self).dump_array(flo,indent,flags)
        if flags['arraylines'] == 0:
            return
        if flags['arraylines'] < 0:
            n = self.nkeys
        else:
            n = min(flags['arraylines'],self.nkeys)
        indent2 = indent+"    "
        for i in xrange(n):
            kf = self.keys[i]
            flo.write("%skeys[0] = %s.Key\n" % (indent,self.label))
            flo.write("%s    frameno = %r\n" % (indent,kf.frameno))
            flo.write("%s    flags = %r\n" % (indent,kf.flags))
            if kf.flags & 1:
                flo.write("%s    tension = %r\n" % (indent,kf.tension))
            if kf.flags & 2:
                flo.write("%s    continuity = %r\n" % (indent,kf.continuity))
            if kf.flags & 4:
                flo.write("%s    bias = %r\n" % (indent,kf.bias))
            if kf.flags & 8:
                flo.write("%s    ease_to = %r\n" % (indent,kf.ease_to))
            if kf.flags & 16:
                flo.write("%s    ease_from = %r\n" % (indent,kf.ease_from))
            for typ,attr in self.keyframe_fields:
                flo.write("%s    %s = %r\n" % (indent,attr,getattr(kf,attr)))
        if n < self.nkeys:
            flo.write("%s    ...\n" % indent)


#
# Magic and Version Chunks
#

class M3DMAGIC(ChunkBase):
    tag = 0x4D4D
    single = "M3D_VERSION version, MDATA mdata, KFDATA kfdata"

class M3D_VERSION(ChunkBase):
    tag = 0x0002
    struct = "long number"

class MDATA(ChunkBase):
    tag = 0x3D3D
    single = ("MESH_VERSION version,"
              "MASTER_SCALE scale,"
              "VIEWPORT_LAYOUT layout,"
              "LO_SHADOW_BIAS lo_shadow_bias,"
              "HI_SHADOW_BIAS hi_shadow_bias,"
              "SHADOW_MAP_SIZE shadow_map_size,"
              "SHADOW_SAMPLES shadow_samples,"
              "SHADOW_RANGE shadow_range,"
              "SHADOW_FILTER shadow_filter,"
              #"RAY_BIAS ray_bias,"
              "O_CONSTS o_consts,"
              "AMBIENT_LIGHT ambient_light,"
              "SOLID_BGND solid_background,"
              "BIT_MAP bitmap,"
              "V_GRADIENT v_gradient,"
              "USE_BIT_MAP use_bitmap,"
              "USE_SOLID_BGND use_solid_background,"
              "USE_V_GRADIENT use_v_gradient,"
              "FOG fog,"
              "LAYER_FOG layer_fog,"
              "DISTANCE_CUE distance_cue,"
              "USE_FOG use_fog,"
              "USE_LAYER_FOG use_layer_fog,"
              "USE_DISTANCE_CUE use_distance_cue,"
              "DEFAULT_VIEW default_view,"
              "MARKER marker")
    multiple = "MAT_ENTRY materials, NAMED_OBJECT objects"

class MESH_VERSION(ChunkBase):
    tag = 0x3D3E
    struct = "long number"

class SMAGIC(UndefinedChunk): tag = 0x2D2D
class LMAGIC(UndefinedChunk): tag = 0x2D3D
class MLIBMAGIC(UndefinedChunk): tag = 0x2DAA
class PRJMAGIC(UndefinedChunk): tag = 0x3D2C
class MATMAGIC(UndefinedChunk): tag = 0x3DFF

#
# Miscellaneous, Global data
#

class MARKER(ChunkBase):
    tag = 0x0001
    struct = "long sum"

class COLOR_F(ChunkBase):
    tag = 0x0010
    struct = "float red, float green, float blue"

class COLOR_24(Color24Chunk):
    tag = 0x0011

class LIN_COLOR_24(Color24Chunk):
    tag = 0x0012

class LIN_COLOR_F(ChunkBase):
    tag = 0x0013
    struct = "float red, float green, float blue"

class INT_PERCENTAGE(OneShortValueChunk):
    tag = 0x0030

class FLOAT_PERCENTAGE(OneFloatValueChunk):
    tag = 0x0031

class MASTER_SCALE(OneFloatValueChunk):
    tag = 0x0100


#
# Coloring, Lighting, Atmosphers Stuff
#

class BIT_MAP(ChunkBase):
    tag = 0x1100
    struct = 'string filename'
    swallow = True

class USE_BIT_MAP(ChunkBase):
    tag = 0x1101

class SOLID_BGND(OneColorChunk):
    tag = 0x1200

class USE_SOLID_BGND(ChunkBase):
    tag = 0x1201

class V_GRADIENT(ChunkBase):
    tag = 0x1300
    struct = "float midpoint"
    multiple = ("COLOR_F color, LIN_COLOR_F lincolor,"
                "COLOR_24 color, LIN_COLOR_24 lincolor")

class USE_V_GRADIENT(ChunkBase):
    tag = 0x1301

class LO_SHADOW_BIAS(OneFloatValueChunk):
    tag = 0x1400

class HI_SHADOW_BIAS(OneFloatValueChunk):
    tag = 0x1410

class SHADOW_MAP_SIZE(OneShortValueChunk):
    tag = 0x1420

class SHADOW_SAMPLES(OneShortValueChunk):
    tag = 0x1430

class SHADOW_RANGE(OneShortValueChunk):
    tag = 0x1440

class SHADOW_FILTER(OneFloatValueChunk):
    tag = 0x1450

class RAW_BIAS(OneFloatValueChunk):
    tag = 0x1460

class O_CONSTS(ChunkBase):
    tag = 0x1500
    struct = "float plane_x, float plane_y, float plane_z"

class AMBIENT_LIGHT(OneColorChunk):
    tag = 0x2100


class FOG(ChunkBase):
    tag = 0x2200
    struct = ("float near_plane, float near_density,"
              "float far_plane, float far_density")
    single = "COLOR_F color, FOG_BGND fog_background"

class USE_FOG(ChunkBase):
    tag = 0x2201

class FOG_BGND(ChunkBase):
    tag = 0x2210

class DISTANCE_CUE(ChunkBase):
    tag = 0x2300
    struct = ("float near_plane, float near_density,"
              "float far_plane, float far_density")
    single = "DCUE_BGND dcue_background"

class USE_DISTANCE_CUE(ChunkBase):
    tag = 0x2301

class LAYER_FOG(ChunkBase):
    tag = 0x2302
    struct = ("float fog_z_from, float fog_z_to,"
              "float fog_density, long fog_type")
    single = "COLOR_F color"

class USE_LAYER_FOG(ChunkBase):
    tag = 0x2303

class DCUE_BGND(ChunkBase):
    tag = 0x2310

#
# View Data
#

class DEFAULT_VIEW(ChunkBase):
    tag = 0x3000
    single = ('VIEW_TOP view, VIEW_BOTTOM view, VIEW_LEFT view,'
              'VIEW_RIGHT view, VIEW_FRONT view, VIEW_BACK view,'
              'VIEW_USER view, VIEW_CAMERA camera')

class ViewChunk(ChunkBase):
    struct = ("float target_x, float target_y,"
              "float target_z, float view_width")

class VIEW_TOP(ViewChunk):
    tag = 0x3010

class VIEW_BOTTOM(ViewChunk):
    tag = 0x3020

class VIEW_LEFT(ViewChunk):
    tag = 0x3030

class VIEW_RIGHT(ViewChunk):
    tag = 0x3040

class VIEW_FRONT(ViewChunk):
    tag = 0x3050

class VIEW_BACK(ViewChunk):
    tag = 0x3060

class VIEW_USER(ChunkBase):
    tag = 0x3070
    struct = ("float target_x, float target_y,"
              "float target_z, float view_width,"
              "float horiz_angle, float vert_angle,"
              "float back_angle")

class VIEW_CAMERA(ChunkBase):
    tag = 0x3080
    struct = "string name"
    swallow = True

class VIEW_WINDOW(UndefinedChunk):
    tag = 0x3090


#
# Objects
#

class NAMED_OBJECT(ChunkBase):
    tag = 0x4000
    struct = "string name"
    single = ("N_TRI_OBJECT obj,"
              "N_DIRECT_LIGHT obj,"
              "N_CAMERA obj,"
              "OBJ_HIDDEN hidden,"
              "OBJ_VIS_LOFTER vis_lofter,"
              "OBJ_DOESNT_CAST doesnt_cast,"
              "OBJ_MATTE matte,"
              "OBJ_DONT_RCVSHADOW dont_rcvshadow,"
              "OBJ_FAST fast,"
              "OBJ_PROCEDURAL procedural,"
              "OBJ_FROZEN frozen")

class OBJ_HIDDEN(ChunkBase):
    tag = 0x4010

class OBJ_VIS_LOFTER(ChunkBase):
    tag = 0x4011

class OBJ_DOESNT_CAST(ChunkBase):
    tag = 0x4012

class OBJ_MATTE(ChunkBase):
    tag = 0x4013

class OBJ_FAST(ChunkBase):
    tag = 0x4014

class OBJ_PROCEDURAL(ChunkBase):
    tag = 0x4015

class OBJ_FROZEN(ChunkBase):
    tag = 0x4016

class OBJ_DONT_RCVSHADOW(ChunkBase):
    tag = 0x4017

class N_TRI_OBJECT(ChunkBase):
    tag = 0x4100
    single = ("POINT_ARRAY points,"
              "POINT_FLAG_ARRAY flags,"
              "FACE_ARRAY faces,"
              "TEX_VERTS texverts,"
              "MESH_MATRIX matrix,"
              "MESH_COLOR color,"
              "MESH_TEXTURE_INFO texinfo,"
              "PROC_NAME proc_name,"
              "PROC_DATA proc_data")
    multiple = "MSH_MAT_GROUP matlist"

class POINT_ARRAY(ArrayChunk):
    tag = 0x4110
    struct = "short npoints"
    def read_array(self,fbuf,flags):
        size = 12*self.npoints
        a = string_to_array(fbuf.read(size),numpy.float32)
        a = numpy.reshape(a,(self.npoints,3))
        self.array = numpy.array(a)
    def dump_array(self,flo,indent,flags):
        super(POINT_ARRAY,self).dump_array(flo,indent,flags)
        if flags['arraylines'] == 0:
            return
        if flags['arraylines'] < 0:
            n = self.npoints
        else:
            n = min(flags['arraylines'],self.npoints)
        for i in xrange(n):
            flo.write("%s    %12.4g%12.4g%12.4g\n"
                      % (indent,self.array[i,0],self.array[i,1],
                         self.array[i,2]))
        if n < self.npoints:
            flo.write("%s            ...\n" % indent)
    def write_array(self):
        s = numpy.array(self.array).astype(numpy.float32)
        return array_to_string_destructive(s)
    swallow = True

class POINT_FLAG_ARRAY(ArrayChunk):
    tag = 0x4111
    struct = "short npoints"
    def read_array(self,fbuf,flags):
        size = 2*self.npoints
        self.array = string_to_array(fbuf.read(size),numpy.uint16)
    def write_array(self):
        s = numpy.array(self.array).astype(numpy.uint16)
        return array_to_string_destructive(s)
    swallow = True

class FACE_ARRAY(ArrayChunk):
    tag = 0x4120
    struct = "short nfaces"
    multiple = 'MSH_MAT_GROUP materials'
    single = 'SMOOTH_GROUP smoothing, MSH_BOXMAP box'
    def read_array(self,fbuf,flags):
        size = 8*self.nfaces
        a = string_to_array(fbuf.read(size),numpy.uint16)
        a = numpy.reshape(a,(self.nfaces,4))
        self.array = numpy.array(a)
    def dump_array(self,flo,indent,flags):
        super(FACE_ARRAY,self).dump_array(flo,indent,flags)
        if flags['arraylines'] == 0:
            return
        if flags['arraylines'] < 0:
            n = self.nfaces
        else:
            n = min(flags['arraylines'],self.nfaces)
        for i in xrange(n):
            flo.write("%s    %10d%10d%10d%10d\n"
                      % (indent,self.array[i,0],self.array[i,1],
                         self.array[i,2],self.array[i,3]))
        if n < self.nfaces:
            flo.write("%s            ...\n" % indent)
    def write_array(self):
        s = numpy.array(self.array).astype(numpy.uint16)
        return array_to_string_destructive(s)

class MSH_MAT_GROUP(ArrayChunk):
    tag = 0x4130
    struct = "string name, short mfaces"
    def read_array(self,fbuf,flags):
        size = 2*self.mfaces
        self.array = string_to_array(fbuf.read(size),numpy.uint16)
    def dump_array(self,flo,indent,flags):
        super(MSH_MAT_GROUP,self).dump_array(flo,indent,flags)
        if flags['arraylines'] == 0:
            return
        if flags['arraylines'] < 0:
            n = self.mfaces
        else:
            n = min(flags['arraylines'],self.mfaces)
        for i in xrange(n):
            flo.write("%s    %10d\n" % (indent,self.array[i]))
        if n < self.mfaces:
            flo.write("%s            ...\n" % indent)
    def write_array(self):
        s = numpy.array(self.array).astype(numpy.uint16)
        return array_to_string_destructive(s)
    swallow = True

class TEX_VERTS(ArrayChunk):
    tag = 0x4140
    struct = "short npoints"
    def read_array(self,fbuf,flags):
        size = 8*self.npoints
        a = string_to_array(fbuf.read(size),numpy.float32)
        a = numpy.reshape(a,(self.npoints,2))
        self.array = numpy.array(a)
    def dump_array(self,flo,indent,flags):
        super(TEX_VERTS,self).dump_array(flo,indent,flags)
        if flags['arraylines'] == 0:
            return
        if flags['arraylines'] < 0:
            n = self.npoints
        else:
            n = min(flags['arraylines'],self.npoints)
        for i in xrange(n):
            flo.write("%s    %12.4g%12.4g\n"
                      % (indent,self.array[i,0],self.array[i,1]))
        if n < self.npoints:
            flo.write("%s            ...\n" % indent)
    def write_array(self):
        s = numpy.array(self.array).astype(numpy.float32)
        return array_to_string_destructive(s)
    swallow = True

class SMOOTH_GROUP(ArrayChunk):
    tag = 0x4150
    def read_array(self,fbuf,flags):
        self.array = string_to_array(fbuf.read_rest(),numpy.uint32)
    def dump_array(self,flo,indent,flags):
        super(SMOOTH_GROUP,self).dump_array(flo,indent,flags)
        if flags['arraylines'] == 0:
            return
        if flags['arraylines'] < 0:
            n = len(self.array)
        else:
            n = min(flags['arraylines'],len(self.array))
        for i in xrange(n):
            flo.write("%s    %10d\n" % (indent,self.array[i]))
        if n < len(self.array):
            flo.write("%s            ...\n" % indent)
    def write_array(self):
        s = numpy.array(self.array).astype(numpy.uint32)
        return array_to_string_destructive(s)
    swallow = True

class MESH_MATRIX(MatrixChunk):
    tag = 0x4160

class MESH_COLOR(ChunkBase):
    tag = 0x4165
    struct = "byte color_index"

class MESH_TEXTURE_INFO(MatrixChunk):
    tag = 0x4170
    __slots__ = [ 'icon_width', 'icon_height', 'cyl_height' ]
    struct = ("float x_tiling, float y_tiling, float icon_x,"
              "float icon_y, float icon_z, float scaling")
    def read(self,fbuf,flags):
        self.read_struct(fbuf,flags)
        self.read_array(fbuf,flags)
        self.icon_width = self.get_float(fbuf,flags)
        self.icon_height = self.get_float(fbuf,flags)
        self.cyl_height = self.get_float(fbuf,flags)
        self.read_chunks(fbuf,flags)
    def dump(self,flo,indent,flags):
        indent += '    '
        self.dump_header(flo,flags)
        self.dump_struct(flo,indent,flags)
        self.dump_array(flo,indent,flags)
        self.out_attr('icon_width',flo,indent,flags)
        self.out_attr('icon_height',flo,indent,flags)
        self.out_attr('cyl_height',flo,indent,flags)
        self.dump_chunks(flo,indent,flags)
    def write(self):
        s = []
        s.append(self.write_struct())
        s.append(self.write_array())
        s.append(self.out_float(self.icon_width))
        s.append(self.out_float(self.icon_height))
        s.append(self.out_float(self.cyl_height))
        s.append(self.write_chunks())
        return ''.join(s)

class PROC_NAME(ChunkBase):
    tag = 0x4181
    struct = "string value"
    swallow = True

class PROC_DATA(UndefinedChunk):
    tag = 0x4181
    struct = "string value"
    swallow = True

class MSH_BOXMAP(ChunkBase):
    tag = 0x4190
    struct = ("string front, string back, string left,"
              "string right, string top, string bottom")
    swallow = True


class N_DIRECT_LIGHT(ChunkBase):
    tag = 0x4600
    struct = "float light_x, float light_y, float light_z"
    single = ("COLOR_F color, COLOR_24 color,"
              "DL_OFF switch,"
              "DL_OUTER_RANGE outer_range,"
              "DL_INNER_RANGE inner_range,"
              "DL_MULTIPLIER multiplier,"
              "DL_SPOTLIGHT spotlight,"
              "DL_ATTENUATE attenuate")
    multiple = "DL_EXCLUDE excludes"

class DL_SPOTLIGHT(ChunkBase):
    tag = 0x4610
    struct = ("float spot_x, float spot_y, float spot_z,"
              "float hotspot_angle, float falloff_angle")
    single = ("DL_SPOT_ROLL roll_angle,"
              "DL_SHADOWED shadowed,"
              "DL_LOCAL_SHADOW2 local_shadow,"
              "DL_SEE_CONE see_cone,"
              "DL_SPOT_RECTANGULAR rectangular,"
              "DL_SPOT_ASPECT aspect,"
              "DL_SPOT_PROJECTOR projector,"
              "DL_SPOT_OVERSHOOT overshoot,"
              "DL_RAY_BIAS bias,"
              "DL_RAYSHADE rayshade")

class DL_OFF(ChunkBase):
    tag = 0x4620

class DL_ATTENUATE(ChunkBase):
    tag = 0x4625

class DL_RAYSHADE(ChunkBase):
    tag = 0x4627

class DL_SHADOWED(ChunkBase):
    tag = 0x4630

class DL_LOCAL_SHADOW2(ChunkBase):
    tag = 0x4630
    struct = "float low_bias, float filter, short mapsize"

class DL_SEE_CONE(ChunkBase):
    tag = 0x4650

class DL_SPOT_RECTANGULAR(ChunkBase):
    tag = 0x4651

class DL_SPOT_OVERSHOOT(ChunkBase):
    tag = 0x4652

class DL_SPOT_PROJECTOR(ChunkBase):
    tag = 0x4653
    struct = "string filename"
    swallow = True

class DL_EXCLUDE(ChunkBase):
    tag = 0x4654
    struct = "string value"
    swallow = True

class DL_SPOT_ROLL(OneFloatValueChunk):
    tag = 0x4656

class DL_SPOT_ASPECT(OneFloatValueChunk):
    tag = 0x4657

class DL_RAY_BIAS(OneFloatValueChunk):
    tag = 0x4658

class DL_INNER_RANGE(OneFloatValueChunk):
    tag = 0x4659

class DL_OUTER_RANGE(OneFloatValueChunk):
    tag = 0x465A

class DL_MULTIPLIER(OneFloatValueChunk):
    tag = 0x465B


class N_CAMERA(ChunkBase):
    tag = 0x4700
    struct = ("float camera_x, float camera_y, float camera_z,"
              "float target_x, float target_y, float target_z,"
              "float bank_angle, float focal_length")
    single = "CAM_SEE_CONE see_cone, CAM_RANGES ranges"

class CAM_SEE_CONE(ChunkBase):
    tag = 0x4710

class CAM_RANGES(ChunkBase):
    tag = 0x4720
    struct = "float near, float far"

#
# Viewport Data
#

class VIEWPORT_LAYOUT(ChunkBase):
    tag = 0x7001
    struct = ('short form, short top, short ready,'
              'short wstate, short swapws, short swapport,'
              'short swapcur')
    single = 'VIEWPORT_SIZE size'
    multiple = 'VIEWPORT_DATA data, VIEWPORT_DATA_3 data'

class ViewportDataChunk(ChunkBase):
    struct = ('short flags, short axis_lockout, short win_x,'
              'short win_y, short win_w, short win_h,'
              'short win_view, float zoom, float worldcenter_x,'
              'float worldcenter_y, float worldcenter_z,'
              'float horiz_ang, float vert_ang,'
              'string cameraname')
    swallow = True

class VIEWPORT_DATA(ViewportDataChunk):
    tag = 0x7011

class VIEWPORT_DATA_3(ViewportDataChunk):
    tag = 0x7012

class VIEWPORT_SIZE(ChunkBase):
    tag = 0x7020
    struct = 'short x, short y, short w, short h'

class NETWORK_VIEW(UndefinedChunk):
    tag = 0x7030

#
# Material Data
#

class MAT_NAME(ChunkBase):
    tag = 0xA000
    struct = 'string value'
    swallow = True

class MAT_AMBIENT(OneColorChunk):
    tag = 0xA010

class MAT_DIFFUSE(OneColorChunk):
    tag = 0xA020

class MAT_SPECULAR(OneColorChunk):
    tag = 0xA030

class MAT_SHININESS(OnePercentageChunk):
    tag = 0xA040

class MAT_SHIN2PCT(OnePercentageChunk):
    tag = 0xA041

class MAT_SHIN3PCT(OnePercentageChunk):
    tag = 0xA042

class MAT_TRANSPARENCY(OnePercentageChunk):
    tag = 0xA050

class MAT_XPFALL(OnePercentageChunk):
    tag = 0xA052

class MAT_REFBLUR(OnePercentageChunk):
    tag = 0xA053

class MAT_SELF_ILLUM(ChunkBase):
    tag = 0xA080

class MAT_TWO_SIDE(ChunkBase):
    tag = 0xA081

class MAT_DECAL(ChunkBase):
    tag = 0xA082

class MAT_ADDITIVE(ChunkBase):
    tag = 0xA083

class MAT_SELF_ILPCT(OnePercentageChunk):
    tag = 0xA084

class MAT_WIRE(ChunkBase):
    tag = 0xA085

class MAT_SUPERSMP(UndefinedChunk):
    tag = 0xA086

class MAT_WIRESIZE(OneFloatValueChunk):
    tag = 0xA087

class MAT_FACEMAP(ChunkBase):
    tag = 0xA088

class MAT_XPFALLIN(UndefinedChunk):
    tag = 0xA08A

class MAT_PHONGSOFT(ChunkBase):
    tag = 0xA08C

class MAT_WIREABS(ChunkBase):
    tag = 0xA08E

class MAT_SHADING(OneShortValueChunk):
    tag = 0xA100

class MAT_TEXMAP(TextureChunk):
    tag = 0xA200

class MAT_SPECMAP(TextureChunk):
    tag = 0xA204

class MAT_OPACMAP(TextureMaskChunk):
    tag = 0xA210

class MAT_REFLMAP(ChunkBase):
    tag = 0xA220
    single = "INT_PERCENTAGE pct, MAT_MAPNAME filename"

class MAT_BUMPMAP(TextureMaskChunk):
    tag = 0xA230

class MAT_USE_XPFALL(ChunkBase):
    tag = 0xA240

class MAT_USE_REFBLUR(ChunkBase):
    tag = 0xA250

class MAT_BUMP_PERCENT(ChunkBase):
    tag = 0xA252
    struct = "short value"

class MAT_MAPNAME(ChunkBase):
    tag = 0xA300
    struct = "string value"
    swallow = True

class MAT_ACUBIC(ChunkBase):
    tag = 0xA310
    struct = ("byte shade, byte antialias, short rflags,"
              "long mapsize, long frame")

class MAT_TEX2MAP(TextureChunk):
    tag = 0xA33A

class MAT_SHINMAP(TextureMaskChunk):
    tag = 0xA33C

class MAT_SELFIMAP(TextureChunk):
    tag = 0xA33D

class MAT_TEXMASK(TextureMaskChunk):
    tag = 0xA33E

class MAT_TEXT2MASK(TextureMaskChunk):
    tag = 0xA340

class MAT_OPACMASK(TextureMaskChunk):
    tag = 0xA342

class MAT_BUMPMASK(TextureMaskChunk):
    tag = 0xA344

class MAT_SHINMASK(TextureMaskChunk):
    tag = 0xA346

class MAT_SPECMASK(TextureMaskChunk):
    tag = 0xA348

class MAT_SELFIMASK(TextureChunk):
    tag = 0xA34A

class MAT_REFLMASK(TextureChunk):
    tag = 0xA34C

class MAT_MAP_TILING(OneShortValueChunk):
    tag = 0xA351

class MAT_MAP_TEXBLUR(OneFloatValueChunk):
    tag = 0xA353

class MAT_MAP_USCALE(OneFloatValueChunk):
    tag = 0xA354

class MAT_MAP_VSCALE(OneFloatValueChunk):
    tag = 0xA356

class MAT_MAP_UOFFSET(OneFloatValueChunk):
    tag = 0xA358

class MAT_MAP_VOFFSET(OneFloatValueChunk):
    tag = 0xA35A

class MAT_MAP_ANG(OneFloatValueChunk):
    tag = 0xA35C

class MAT_MAP_COL1(Color24Chunk):
    tag = 0xA360

class MAT_MAP_COL2(Color24Chunk):
    tag = 0xA362

class MAT_MAP_RCOL(Color24Chunk):
    tag = 0xA364

class MAT_MAP_GCOL(Color24Chunk):
    tag = 0xA366

class MAT_MAP_BCOL(Color24Chunk):
    tag = 0xA368

class MAT_ENTRY(ChunkBase):
    tag = 0xAFFF
    single = ("MAT_NAME name,"
              "MAT_AMBIENT ambient,"
              "MAT_DIFFUSE diffuse,"
              "MAT_SPECULAR specular,"
              "MAT_SHININESS shininess,"
              "MAT_SHIN2PCT shin2pct,"
              "MAT_TRANSPARENCY transparency,"
              "MAT_XPFALL xpfall,"
              "MAT_REFBLUR refblur,"
              "MAT_SELF_ILLUM self_illum,"
              "MAT_SHADING shading,"
              "MAT_SELF_ILPCT self_ilpct,"
              "MAT_USE_XPFALL use_xpfall,"
              "MAT_USE_REFBLUR use_refblur,"
              "MAT_TWO_SIDE two_side,"
              "MAT_ADDITIVE additive,"
              "MAT_WIRE wire,"
              "MAT_FACEMAP facemap,"
              "MAT_PHONGSOFT phongsoft,"
              "MAT_WIRESIZE wiresize,"
              "MAT_DECAL decal,"
              "MAT_TEXMAP texmap,"
              "MAT_SXP_TEXT_DATA sxp_text_data,"
              "MAT_TEXMASK texmask,"
              "MAT_SXP_TEXT_MASKDATA sxp_text_maskdata,"
              "MAT_TEX2MAP tex2map,"
              "MAT_SXP_TEXT2_DATA sxp_text2_data,"
              "MAT_TEXT2MASK text2mask,"
              "MAT_SXP_TEXT2_MASKDATA sxp_text2_maskdata,"
              "MAT_OPACMAP opacmap,"
              "MAT_SXP_OPAC_DATA sxp_opac_data,"
              "MAT_OPACMASK opac_mask,"
              "MAT_SXP_OPAC_MASKDATA sxp_opac_maskdata,"
              "MAT_BUMPMAP bumpmap,"
              "MAT_SXP_BUMP_DATA sxp_bump_data,"
              "MAT_BUMPMASK bumpmask,"
              "MAT_SXP_BUMP_MASKDATA sxp_bump_maskdata,"
              "MAT_SPECMAP specmap,"
              "MAT_SXP_SPEC_DATA sxp_spec_data,"
              "MAT_SPECMASK specmask,"
              "MAT_SXP_SPEC_MASKDATA sxp_spec_maskdata,"
              "MAT_SHINMAP shinmap,"
              "MAT_SXP_SHIN_DATA sxp_shin_data,"
              "MAT_SHINMASK shinmask,"
              "MAT_SXP_SHIN_MASKDATA sxp_shin_maskdata,"
              "MAT_SELFIMAP selfimap,"
              "MAT_SXP_SELFI_DATA sxp_selfi_data,"
              "MAT_SELFIMASK selfimask,"
              "MAT_SXP_SELFI_MASKDATA sxp_selfi_maskdata,"
              "MAT_REFLMAP reflmap,"
              "MAT_REFLMASK reflmask,"
              "MAT_SXP_REFL_MASKDATA sxp_refl_maskdata,"
              "MAT_ACUBIC acubic")

class MAT_SXP_TEXT_DATA(UndefinedChunk): tag = 0xA320
class MAT_SXP_TEXT2_DATA(UndefinedChunk): tag = 0xA321
class MAT_SXP_OPAC_DATA(UndefinedChunk): tag = 0xA322
class MAT_SXP_BUMP_DATA(UndefinedChunk): tag = 0xA324
class MAT_SXP_SPEC_DATA(UndefinedChunk): tag = 0xA325
class MAT_SXP_SHIN_DATA(UndefinedChunk): tag = 0xA326
class MAT_SXP_SELFI_DATA(UndefinedChunk): tag = 0xA328
class MAT_SXP_TEXT_MASKDATA(UndefinedChunk): tag = 0xA32A
class MAT_SXP_TEXT2_MASKDATA(UndefinedChunk): tag = 0xA32C
class MAT_SXP_OPAC_MASKDATA(UndefinedChunk): tag = 0xA32E
class MAT_SXP_BUMP_MASKDATA(UndefinedChunk): tag = 0xA330
class MAT_SXP_SPEC_MASKDATA(UndefinedChunk): tag = 0xA332
class MAT_SXP_SHIN_MASKDATA(UndefinedChunk): tag = 0xA334
class MAT_SXP_SELFI_MASKDATA(UndefinedChunk): tag = 0xA336
class MAT_SXP_REFL_MASKDATA(UndefinedChunk): tag = 0xA338

#
# Keyframe Section
#

class KFDATA(ChunkBase):
    tag = 0xB000
    single = "KFHDR kfhdr, KFSEG kfseg, KFCURTIME kfcurtime"
    multiple = ("OBJECT_NODE_TAG object_nodes,"
                "CAMERA_NODE_TAG camera_nodes,"
                "TARGET_NODE_TAG target_nodes,"
                "LIGHT_NODE_TAG light_nodes,"
                "SPOTLIGHT_NODE_TAG spotlight_nodes,"
                "L_TARGET_NODE_TAG l_target_nodes,"
                "AMBIENT_NODE_TAG ambient_nodes")

class KFHDR(ChunkBase):
    tag = 0xB00A
    struct = "short revision, string filename, long anim_length"

class KFSEG(ChunkBase):
    tag = 0xB008
    struct = "long first_frame, long last_frame"

class KFCURTIME(ChunkBase):
    tag = 0xB009
    struct = "long current_frame"

class AMBIENT_NODE_TAG(ChunkBase):
    tag = 0xB001
    single = "NODE_ID node_id, NODE_HDR node_hdr, COL_TRACK_TAG col_track"

class OBJECT_NODE_TAG(ChunkBase):
    tag = 0xB002
    single = ("NODE_ID node_id,"
              "NODE_HDR node_hdr,"
              "PIVOT pivot,"
              "INSTANCE_NAME instance_name,"
              "BOUNDBOX bounding_box,"
              "POS_TRACK_TAG pos_track,"
              "ROT_TRACK_TAG rot_track,"
              "SCL_TRACK_TAG scl_track,"
              "MORPH_TRACK_TAG morph_track,"
              "HIDE_TRACK_TAG hide_track,"
              "MORPH_SMOOTH morph_smooth")


class CAMERA_NODE_TAG(ChunkBase):
    tag = 0xB003
    single = ("NODE_ID node_id,"
              "NODE_HDR node_hdr,"
              "POS_TRACK_TAG pos_track,"
              "FOV_TRACK_TAG fov_track,"
              "ROLL_TRACK_TAG roll_track")

class TARGET_NODE_TAG(ChunkBase):
    tag = 0xB004    
    single = ("NODE_ID node_id,"
              "NODE_HDR node_hdr,"
              "POS_TRACK_TAG pos_track")


class LIGHT_NODE_TAG(ChunkBase):
    tag = 0xB005
    single = ("NODE_ID node_id,"
              "NODE_HDR node_hdr,"
              "POS_TRACK_TAG pos_track,"
              "COL_TRACK_TAG col_track")

class SPOTLIGHT_NODE_TAG(ChunkBase):
    tag = 0xB007
    single = ("NODE_ID node_id,"
              "NODE_HDR node_hdr,"
              "POS_TRACK_TAG pos_track,"
              "COL_TRACK_TAG col_track,"
              "HOT_TRACK_TAG hot_track,"
              "FALL_TRACK_TAG fall_track,"
              "ROLL_TRACK_TAG roll_track")

class L_TARGET_NODE_TAG(ChunkBase):
    tag = 0xB006 
    single = ("NODE_ID node_id,"
              "NODE_HDR node_hdr,"
              "POS_TRACK_TAG pos_track,"
              "COL_TRACK_TAG col_track")

class NODE_ID(ChunkBase):
    tag = 0xB030
    struct = "short id"

class NODE_HDR(ChunkBase):
    tag = 0xB010
    struct = "string name, short flags1, short flags2, short parent"

class PIVOT(ChunkBase):
    tag = 0xB013
    struct = "float pivot_x, float pivot_y, float pivot_z"

class INSTANCE_NAME(ChunkBase):
    tag = 0xB011
    struct = "string name"

class MORPH_SMOOTH(ChunkBase):
    tag = 0xB015
    struct = "float smooth_angle"

class BOUNDBOX(ChunkBase):
    tag = 0xB014
    struct = ("float min_x, float min_y, float min_z,"
              "float max_x, float max_y, float max_z")

class POS_TRACK_TAG(TrackChunk):
    tag = 0xB020
    keyframe = "float pos_x, float pos_y, float pos_z"

class COL_TRACK_TAG(TrackChunk):
    tag = 0xB025
    keyframe = "float red, float green, float blue"

class ROT_TRACK_TAG(TrackChunk):
    tag = 0xB021
    keyframe = "float angle, float axis_x, float axis_y, float axis_z"

class SCL_TRACK_TAG(TrackChunk):
    tag = 0xB022
    keyframe = "float scl_x, float scl_y, float scl_z"

class MORPH_TRACK_TAG(TrackChunk):
    tag = 0xB026
    keyframe = "string name"

class FOV_TRACK_TAG(TrackChunk):
    tag = 0xB023
    keyframe = "float angle"

class ROLL_TRACK_TAG(TrackChunk):
    tag = 0xB024
    keyframe = "float angle"

class HOT_TRACK_TAG(TrackChunk):
    tag = 0xB027
    keyframe = "float angle"

class FALL_TRACK_TAG(TrackChunk):
    tag = 0xB028
    keyframe = "float angle"

class HIDE_TRACK_TAG(UndefinedChunk):
    tag = 0xB029

#
# Misc
#

class DUMMY(ChunkBase): tag = 0xFFFF

#
# Obsolute Chunks
#

class VIEWPORT_LAYOUT_OLD(UndefinedChunk): tag = 0x7000
class VIEWPORT_DATA_OLD(UndefinedChunk): tag = 0x7010
class OLD_MAT_GROUP(UndefinedChunk): tag = 0x4131
class MAT_MAP_TILING_OLD(UndefinedChunk): tag = 0xA350
class MAT_MAP_TEXBLUR_OLD(UndefinedChunk): tag = 0xA352


#
# Functions to operate on chunks
#

def read_3ds_mem(membuf,check_magic=True,tight=False,recover=True):
    """Create a 3DS DOM from a memory buffer.

        dom = read_3ds_mem(buffer,check_magic=True,tight=False,
                           recover=True)

    buffer: is an image of the 3DS file in memory.  It could be
    a string, a mapped file, or something else.

    check_magic: If true, this function checks that the top level
    chunk is the 3DS magic chunk (0x4D4D), and raises an exception
    if it is not.

    tight: Whether to use tighter error checking.  Try disabling
    if getting 3DS format errors.

    recover: Whether to emit an Error chunk when an error is found;
    otherwise raise an exception.

    """

    if check_magic:
        tag,length = struct.unpack("<HL",membuf[:6])
        if tag != 0x4D4D:
            raise File3dsFormatError("Not a 3D Studio file.")
    fbuf = FileLikeBuffer(membuf,0,len(membuf))
    flags = { 'tight': tight, 'recover': recover }
    return ChunkBase.get_chunk(fbuf,flags)


def read_3ds_file(filename,check_magic=True,tight=False,recover=True):
    """Create a 3DS DOM from a file.

        dom = read_3ds_file(filename,check_magic=True,tight=False,
                            recover=True)

    filename: name of a 3DS file.

    check_magic: If true, this function checks that the top level
    chunk is the 3DS magic chunk (0x4D4D), and raises an exception
    if it is not.

    tight: Whether to use tighter error checking.  Try disabling
    if getting 3DS format errors.

    recover: Whether to emit an Error chunk when an error is found;
    otherwise raise an exception.

    """

    flo = open(filename,'rb')
    try:
        return read_3ds_mem(flo.read(),check_magic,tight,recover)
    finally:
        flo.close()


def write_3ds_mem(dom,check_magic=True):
    """Output a 3DS DOM as a string.

        buf = write_3ds_mem(dom,check_magic=True)

    dom: the 3DS dom

    check_magic: If true, this function checks that the top level
    chunk is the 3DS magic chunk (0x4D4D), and raises an exception
    if it is not.

    """

    if check_magic and dom.tag != 0x4D4D:
        raise File3dsFormatError("Not a 3D Studio file.")
    s = dom.write()
    length = len(s) + 6
    return struct.pack("<HL",dom.tag,length) + s


def write_3ds_file(filename,dom,check_magic=True):
    """Write a 3DS file.

        buf = write_3ds_file(filename,dom,check_magic=True)

    filename: name of a 3DS file to write.

    dom: the 3DS dom

    check_magic: If true, this function checks that the top level
    chunk is the 3DS magic chunk (0x4D4D), and raises an exception
    if it is not.

    """

    if check_magic and dom.tag != 0x4D4D:
        raise File3dsFormatError("Not a 3D Studio file.")
    flo = open(filename,'wb')
    try:
        flo.write(write_3ds_mem(dom,False))
    finally:
        flo.close()


def dump_3ds_chunk(chunk,flo,arraylines=10,indent=''):
    """Dump a 3DS DOM to a file stream.

        dump_3ds_chunk(filename,flo,arraylines=10,indent='')

    chunk: The 3DS chunk to dump

    flo: The file-like-object to dump output to (for example, sys.stdout)

    arraylines: Max number of lines of array data to dump. If negative,
    dump the whole array.

    indent: Prefix string to all lines dumped; used to indent.

    """

    flags = { 'arraylines': arraylines }
    chunk.dump(flo,indent,flags)


def dump_3ds_file(filename,flo,arraylines=10,tight=False,recover=True):
    """Dump a text representation of 3DS DOM to a file stream.

        dump_3ds_file(filename,flo,arraylines=2,tight=False,
                      recover=True)

    filename: The 3DS file to dump

    flo: The file-like-object to dump output to (for example, sys.stdout)

    arraylines: Max number of lines of array data to dump. If negative,
    dump the whole array.

    tight: Whether to use tighter error checking.  Try disabling
    if getting 3DS format errors.

    recover: Whether to emit an Error chunk when an error is found;
    otherwise raise an exception.

    """
    
    check_magic = False
    dom = read_3ds_file(filename,check_magic,tight,recover)
    dump_3ds_chunk(dom,flo,arraylines)


def remove_errant_chunks(chunk):
    """Recursively remove any errant chunks.

        remove_errant_chunks(chunk)

    This recursively removes chunks that might prevent
    successfully writing the DOM.

    """

    for attr in chunk.single_order:
        c = getattr(chunk,attr)
        if isinstance(c,ErrorChunk):
            setattr(chunk,attr,None)
        elif c:
            remove_errant_chunks(c)
    mchunks = chunk.multiple_order[:]
    if chunk.freechunks:
        mchunks.append('subchunks')
    for attr in mchunks:
        cl = [ x for x in getattr(chunk,attr) if not isinstance(x,ErrorChunk) ]
        setattr(chunk,attr,cl)
        for c in cl:
            remove_errant_chunks(c)





# -*- coding: utf-8 -*-

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path

from PathTests.PathTestUtils import PathTestBase

class TestPathTooltable(PathTestBase):

    def test00(self):
        '''Verify templateAttrs'''

        t = Path.Tool(name='t', diameter=1.2)
        u = Path.Tool(name='u', diameter=3.4)
        v = Path.Tool(name='v', diameter=5.6)

        tt = Path.Tooltable()
        tt.setTool(3, t)
        tt.setTool(1, u)
        tt.addTools(v)

        attrs = tt.templateAttrs()
        self.assertEqual(3, len(attrs))
        self.assertTrue(1 in attrs)
        self.assertFalse(2 in attrs)
        self.assertTrue(3 in attrs)
        self.assertTrue(4 in attrs)

        self.assertEqual(attrs[1]['name'], 'u')
        self.assertEqual(attrs[1]['diameter'], 3.4)
        self.assertEqual(attrs[3]['name'], 't')
        self.assertEqual(attrs[3]['diameter'], 1.2)
        self.assertEqual(attrs[4]['name'], 'v')
        self.assertEqual(attrs[4]['diameter'], 5.6)
        return tt

    def test01(self):
        '''Verify setFromTemplate roundtrip.'''
        tt = self.test00()
        uu = Path.Tooltable()
        uu.setFromTemplate(tt.templateAttrs())

        self.assertEqual(tt.Content, uu.Content)


    def test02(self):
        '''Verify template constructor.'''
        tt = self.test00()
        uu = Path.Tooltable(tt.templateAttrs())
        self.assertEqual(tt.Content, uu.Content)

